/* Local variables and the run time stack. */
/* #includes */ /*{{{C}}}*//*{{{*/
#undef  _POSIX_SOURCE
#define _POSIX_SOURCE   1
#undef  _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 2

#include "config.h"

#include <assert.h>
#include <ctype.h>
#include <stdlib.h>

#include "auto.h"

#ifdef USE_DMALLOC
#include "dmalloc.h"
#endif
/*}}}*/
/* #defines */ /*{{{*/
#define INCREASE_STACK 16
/*}}}*/

/* interpretation methods */
struct Auto *Auto_new(struct Auto *this) /*{{{*/
{
  this->stackPointer=0;
  this->stackCapacity=0;
  this->framePointer=0;
  this->frameSize=0;
  this->onerror.line=-1;
  this->erl=0;
  Value_new_NIL(&this->err);
  this->slot=(union AutoSlot*)0;
  this->symCapacity=0;
  this->symSize=0;
  this->sym=(struct AutoSym*)0;
  return this;
}
/*}}}*/
void Auto_destroy(struct Auto *this) /*{{{*/
{
  Value_destroy(&this->err);
  if (this->stackCapacity) free(this->slot);
  if (this->symCapacity) free(this->sym);
}
/*}}}*/
struct Var *Auto_pushArg(struct Auto *this) /*{{{*/
{
  if ((this->stackPointer+1)>=this->stackCapacity)
  {
    this->slot=realloc(this->slot,sizeof(this->slot[0])*(this->stackCapacity?(this->stackCapacity=this->stackPointer+INCREASE_STACK):(this->stackCapacity=INCREASE_STACK)));
  }
  return &this->slot[this->stackPointer++].var;
}
/*}}}*/
void Auto_pushFuncRet(struct Auto *this, int firstarg, struct Pc *pc) /*{{{*/
{
  if (this->stackPointer+2>=this->stackCapacity)
  {
    this->slot=realloc(this->slot,sizeof(this->slot[0])*(this->stackCapacity?(this->stackCapacity=this->stackCapacity+INCREASE_STACK):(this->stackCapacity=INCREASE_STACK)));
  }
  this->slot[this->stackPointer].retException.onerror=this->onerror;
  this->slot[this->stackPointer].retException.resumeable=this->resumeable;
  ++this->stackPointer;
  this->slot[this->stackPointer].retFrame.pc=*pc;
  this->slot[this->stackPointer].retFrame.framePointer=this->framePointer;
  this->slot[this->stackPointer].retFrame.frameSize=this->frameSize;
  ++this->stackPointer;
  this->framePointer=firstarg;
  this->frameSize=this->stackPointer-firstarg;
  this->onerror.line=-1;
}
/*}}}*/
void Auto_pushGosubRet(struct Auto *this, struct Pc *pc) /*{{{*/
{
  if ((this->stackPointer+1)>=this->stackCapacity)
  {
    this->slot=realloc(this->slot,sizeof(this->slot[0])*(this->stackCapacity?(this->stackCapacity=this->stackPointer+INCREASE_STACK):(this->stackCapacity=INCREASE_STACK)));
  }
  this->slot[this->stackPointer].retFrame.pc=*pc;
  ++this->stackPointer;
}
/*}}}*/
struct Var *Auto_local(struct Auto *this, int l) /*{{{*/
{
  assert(this->frameSize>l);
  return &(this->slot[this->framePointer+l].var);
}
/*}}}*/
int Auto_funcReturn(struct Auto *this, struct Pc *pc) /*{{{*/
{
  int i,retFrame,retException;

  if (this->stackPointer==0) return 0;
  assert(this->frameSize);
  retFrame=this->framePointer+this->frameSize-1;
  retException=this->framePointer+this->frameSize-2;
  assert(retException>=0 && retFrame<this->stackPointer);
  for (i=0; i<this->frameSize-2; ++i) Var_destroy(&this->slot[this->framePointer+i].var);
  this->stackPointer=this->framePointer;
  if (pc!=(struct Pc*)0) *pc=this->slot[retFrame].retFrame.pc;
  this->frameSize=this->slot[retFrame].retFrame.frameSize;
  this->framePointer=this->slot[retFrame].retFrame.framePointer;
  this->onerror=this->slot[retException].retException.onerror;
  return 1;
}
/*}}}*/
int Auto_gosubReturn(struct Auto *this, struct Pc *pc) /*{{{*/
{
  if (this->stackPointer<=this->framePointer+this->frameSize) return 0;
  --this->stackPointer;
  if (pc) *pc=this->slot[this->stackPointer].retFrame.pc;
  return 1;
}
/*}}}*/
void Auto_frameToError(struct Auto *this, struct Program *program, struct Value *v) /*{{{*/
{
  int i=this->stackPointer,framePointer,frameSize,retFrame;
  struct Pc p;

  framePointer=this->framePointer;
  frameSize=this->frameSize;
  while (i>framePointer+frameSize)
  {
    p=this->slot[--i].retFrame.pc;
    Value_errorSuffix(v,"Called");
    Program_PCtoError(program,&p,v);
  }
  if (i)
  {
    retFrame=framePointer+frameSize-1;
    i=framePointer;
    p=this->slot[retFrame].retFrame.pc;
    frameSize=this->slot[retFrame].retFrame.frameSize;
    framePointer=this->slot[retFrame].retFrame.framePointer;
    Value_errorSuffix(v,"Proc Called");
    Program_PCtoError(program,&p,v);
  }
}
/*}}}*/
void Auto_setError(struct Auto *this, long int line, struct Pc *pc, struct Value *v) /*{{{*/
{
  this->erpc=*pc;
  this->erl=line;
  Value_destroy(&this->err);
  Value_clone(&this->err,v);
}
/*}}}*/

/* compilation methods */
int Auto_find(struct Auto *this, struct Identifier *ident) /*{{{*/
{
  int i;

  for (i=0; i<this->symSize; ++i)
  {
    const char *s=ident->name;
    const char *r=this->sym[i].identifier;

    while (*s && tolower(*s)==tolower(*r)) { ++s; ++r; };
    if (tolower(*s)==tolower(*r))
    {
      ident->sym=&this->sym[i].sym;
      return 1;
    }
  }
  return 0;
}
/*}}}*/
int Auto_variable(struct Auto *this, const struct Identifier *ident) /*{{{*/
{
  int i;

  for (i=0; i<this->symSize; ++i)
  {
    const char *s=ident->name;
    const char *r=this->sym[i].identifier;

    while (*s && tolower(*s)==tolower(*r)) { ++s; ++r; };
    if (tolower(*s)==tolower(*r)) return 0;
  }
  if (this->symSize==this->symCapacity)
  {
    this->sym=realloc(this->sym,sizeof(struct AutoSym)*(this->symCapacity?(this->symCapacity*=2):(this->symCapacity=16)));
  }
  this->sym[this->symSize].identifier=ident->name;
  this->sym[this->symSize].type=ident->defaultType;
  this->sym[this->symSize].sym.type=LOCALVAR;
  this->sym[this->symSize].sym.u.offset=this->symSize;
  ++this->symSize;
  return 1;
}
/*}}}*/
enum ValueType Auto_argType(const struct Auto *this, int l) /*{{{*/
{
  return this->sym[this->symSize-1-l].type;
}
/*}}}*/
void Auto_funcEnd(struct Auto *this) /*{{{*/
{
  this->symSize=0;
}
/*}}}*/
