/* #includes */ /*{{{C}}}*//*{{{*/
#undef  _POSIX_SOURCE
#define _POSIX_SOURCE   1
#undef  _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 2

#include "config.h"

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "getopt.h"

#include "auto.h"
#include "bas.h"
#include "error.h"
#include "fs.h"
#include "global.h"
#include "program.h"
#include "value.h"
#include "var.h"

#ifdef USE_DMALLOC
#include "dmalloc.h"
#endif
/*}}}*/
/* #defines */ /*{{{*/
#define STDCHANNEL 0
#define LPCHANNEL 32
#define DIRECTMODE (pc.line==-1)
#ifndef __GNUC__
#define inline
#endif
/*}}}*/

/* types */ /*{{{*/
enum LabelType
{
  L_IF=1,
  L_ELSE,
  L_DO,
  L_DOWHILE,
  L_FOR,
  L_FOR_VAR,
  L_FOR_LIMIT,
  L_FOR_BODY,
  L_REPEAT,
  L_SELECTCASE,
  L_WHILE,
  L_FUNC
};

struct LabelStack
{
  enum LabelType type;
  struct Pc patch;
};
/*}}}*/
/* variables */ /*{{{*/
static unsigned int labelStackPointer,labelStackCapacity;
static struct LabelStack *labelStack;
static struct Pc *lastdata;
static struct Pc begindata;
static struct Pc curdata;
static struct Pc nextdata;
static enum { DECLARE, COMPILE, INTERPRET } pass;
static int stop;
static int runnable;
static struct Pc pc;
static struct Auto stack;
static struct Program program;
static struct Global globals;
/*}}}*/
/* forward prototypes */ /*{{{*/
static struct Value *statements(struct Value *value);
static struct Value *compile(struct Value *v);
static struct Value *eval(struct Value *value, const char *desc);
static void allowBreak(int on);
/*}}}*/

static struct Value *lvalue(struct Value *value) /*{{{*/
{
  struct Symbol *sym;
  struct Pc lvpc=pc;

  sym=pc.token->u.identifier->sym;
  assert(pass==DECLARE || sym->type==GLOBALVAR || sym->type==LOCALVAR);
  if ((pc.token+1)->type==T_OP)
  {
    struct Pc idxpc;
    unsigned int dim,capacity,*idx;

    pc.token+=2;
    dim=0;
    capacity=0;
    idx=(unsigned int*)0;
    while (1)
    {
      if (pass==INTERPRET && dim==capacity) /* enlarge idx */ /*{{{*/
      {
        idx=realloc(idx,sizeof(unsigned int)*(capacity?(capacity*=2):(capacity=3)));
      }
      /*}}}*/
      idxpc=pc;
      if (eval(value,"index")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR)
      {
        if (capacity) free(idx);
        pc=idxpc;
        return value;
      }
      if (value->type!=V_ERROR) Value_retype(value,V_INTEGER);
      if (pass==INTERPRET)
      {
        if (value->type!=V_ERROR) idx[dim]=value->u.integer;
        ++dim;
      }
      if (value->type==V_ERROR) /* abort */ /*{{{*/
      {
        pc=idxpc;
        free(idx);
        return value;
      }
      /*}}}*/
      if (pc.token->type==T_COMMA)
      {
        ++pc.token;
        Value_destroy(value);
      }
      else break;
    }
    if (pc.token->type!=T_CP)
    {
      if (pass==INTERPRET) free(idx);
      return Value_new_ERROR(value,MISSINGCP);
    }
    else ++pc.token;
    if (pass==INTERPRET)
    {
      if ((value=Var_value(&(sym->u.var),dim,idx,value))->type==V_ERROR) pc=lvpc;
      free(idx);
      return value;
    }
    if (pass==DECLARE)
    {
      return Value_nullValue(V_INTEGER);
    }
    if (pass==COMPILE)
    {
      return Value_nullValue(sym->type==GLOBALVAR ? sym->u.var.type : Auto_local(&stack,sym->u.offset)->type);
    }
    assert(0);
    return (struct Value*)0;
  }
  else
  {
    if (pass==INTERPRET)
    {
      if ((value=Var_value(sym->type==GLOBALVAR ? &(sym->u.var) : Auto_local(&stack,sym->u.offset),0,(unsigned int*)0,value))->type==V_ERROR) return value;
      ++pc.token;
      return value;
    }
    if (pass==DECLARE)
    {
      ++pc.token;
      return Value_nullValue(V_INTEGER);
    }
    if (pass==COMPILE)
    {
      ++pc.token;
      return Value_nullValue(sym->type==GLOBALVAR ? sym->u.var.type : stack.sym[sym->u.offset].type);
    }
    assert(0);
    return (struct Value*)0;
  }
}
/*}}}*/
static struct Value *func(struct Value *value) /*{{{*/
{
  struct Identifier *ident;
  struct Pc funcpc=pc;
  int firstslot=-99;
  int args=0;
  int argerr;
  struct Symbol *sym;
  int overloaded;

  assert(pc.token->type==T_IDENTIFIER);
  ident=pc.token->u.identifier;
  assert(pass==DECLARE || ident->sym->type==BUILTINFUNCTION || ident->sym->type==USERFUNCTION);
  ++pc.token;
  if (pass!=DECLARE)
  {
    firstslot=stack.stackPointer;
    if (ident->sym->type==USERFUNCTION && ident->sym->u.sub.retType!=V_VOID)
    {
      struct Var *v=Auto_pushArg(&stack);
      Var_new(v,ident->sym->u.sub.retType,0,(const unsigned int*)0);
    }
  }
  if (pc.token->type==T_OP) /* push arguments to stack */ /*{{{*/
  {
    ++pc.token;
    if (pc.token->type!=T_CP) while (1)
    {
      if (pass==DECLARE)
      {
        if (eval(value,"actual parameter")->type==V_ERROR) return value;
        Value_destroy(value);
      }
      else
      {
        struct Var *v=Auto_pushArg(&stack);

        Var_new_scalar(v);
        if (eval(v->value,(const char*)0)->type==V_ERROR)
        {
          Value_clone(value,v->value);
          while (stack.stackPointer>firstslot) Var_destroy(&stack.slot[--stack.stackPointer].var);
          return value;
        }
        v->type=v->value->type;
      }
      ++args;
      if (pc.token->type==T_COMMA) ++pc.token;
      else break;
    }
    if (pc.token->type!=T_CP)
    {
      if (pass!=DECLARE)
      {
        while (stack.stackPointer>firstslot) Var_destroy(&stack.slot[--stack.stackPointer].var);
      }
      return Value_new_ERROR(value,MISSINGCP);
    }
    ++pc.token;
  }
  /*}}}*/
  if (pass==DECLARE) Value_new_null(value,ident->defaultType);
  else
  {
    int i;
    int nomore;

    if (pass==INTERPRET && ident->sym->type==USERFUNCTION)
    {
      for (i=0; i<ident->sym->u.sub.u.def.localLength; ++i)
      {
        struct Var *v=Auto_pushArg(&stack);
        Var_new(v,ident->sym->u.sub.u.def.localTypes[i],0,(const unsigned int*)0);
      }
    }
    Auto_pushFuncRet(&stack,firstslot,&pc);

    sym=ident->sym;
    overloaded=(pass==COMPILE && sym->type==BUILTINFUNCTION && sym->u.sub.u.bltin.next);
    do
    {
      nomore=(pass==COMPILE && !(sym->type==BUILTINFUNCTION && sym->u.sub.u.bltin.next));
      argerr=0;
      if (args<sym->u.sub.argLength) /*{{{*/
      {
        if (nomore) Value_new_ERROR(value,TOOFEW);
        argerr=1;
      }
      /*}}}*/
      else if (args>sym->u.sub.argLength) /*{{{*/
      {
        if (nomore) Value_new_ERROR(value,TOOMANY);
        argerr=1;
      }
      /*}}}*/
      else /*{{{*/
      {
        for (i=0; i<args; ++i)
        {
          struct Value *arg=Var_value(Auto_local(&stack,i),0,(unsigned int*)0,value);

          assert(arg->type!=V_ERROR);
          if (overloaded)
          {
            if (arg->type!=sym->u.sub.argTypes[i])
            {
              if (nomore) Value_new_ERROR(value,TYPEMISMATCH2,i+1);
              argerr=1;
              break;
            }
          }
          else if (Value_retype(arg,sym->u.sub.argTypes[i])->type==V_ERROR)
          {
            if (nomore) Value_new_ERROR(value,TYPEMISMATCH3,arg->u.error,i+1);
            argerr=1;
            break;
          }
        }
      }
      /*}}}*/
      if (argerr)
      {
        if (nomore)
        {
          Auto_funcReturn(&stack,(struct Pc*)0);
          pc=funcpc;
          return value;
        }
        else sym=sym->u.sub.u.bltin.next;
      }
    } while (argerr);
    ident->sym=sym;
    if (sym->type==BUILTINFUNCTION)
    {
      if (pass==INTERPRET)
      {
        if (sym->u.sub.u.bltin.call(value,&stack)->type==V_ERROR) pc=funcpc;
      }
      else Value_new_null(value,sym->u.sub.retType);
    }
    else if (sym->type==USERFUNCTION)
    {
      if (pass==INTERPRET)
      {
        int r=1;

        pc=sym->u.sub.u.def.scope.start;
        if (pc.token->type==T_COLON) ++pc.token;
        else Program_skipEOL(&program,&pc,STDCHANNEL,1);
        do
        {
          if (statements(value)->type==V_ERROR)
          {
            if (strchr(value->u.error.msg,'\n')==(char*)0)
            {
              Auto_setError(&stack,Program_lineNumber(&program,&pc),&pc,value);
              Program_PCtoError(&program,&pc,value);
            }
            if (stack.onerror.line!=-1)
            {
              stack.resumeable=1;
              pc=stack.onerror;
            }
            else
            {
              Auto_frameToError(&stack,&program,value);
              break;
            }
          }
          else if (value->type!=V_NIL) break;
          Value_destroy(value);
        } while (r=Program_skipEOL(&program,&pc,STDCHANNEL,1));
        if (!r) Value_new_VOID(value);
      }
      else Value_new_null(value,sym->u.sub.retType);
    }
    Auto_funcReturn(&stack,pass==INTERPRET && value->type!=V_ERROR ? &pc : (struct Pc*)0);
  }
  return value;
}
/*}}}*/

#ifdef LR0
/* Grammar with LR(0) sets */ /*{{{*/
/*
Grammar:

1 EV -> E
2 E  -> E op E
3 E  -> op E
4 E  -> ( E )
5 E  -> value

i0:
EV -> . E                goto(0,E)=5
E  -> . E op E           goto(0,E)=5
E  -> . op E      +,-    shift 2
E  -> . ( E )     (      shift 3
E  -> . value     value  shift 4

i5:
EV -> E .         else   accept
E  -> E . op E    op     shift 1

i2:
E  -> op . E             goto(2,E)=6
E  -> . E op E           goto(2,E)=6
E  -> . op E      +,-    shift 2
E  -> . ( E )     (      shift 3
E  -> . value     value  shift 4

i3:
E  -> ( . E )            goto(3,E)=7
E  -> . E op E           goto(3,E)=7
E  -> . op E      +,-    shift 2
E  -> . ( E )     (      shift 3
E  -> . value     value  shift 4

i4:
E  -> value .            reduce 5

i1:
E  -> E op . E           goto(1,E)=8
E  -> . E op E           goto(1,E)=8
E  -> . op E      +,-    shift 2
E  -> . ( E )     (      shift 3
E  -> . value     value  shift 4

i6:
E  -> op E .             reduce 3
E  -> E . op E    op*    shift 1 *=if stack[-2] contains op of unary lower priority

i7:
E  -> ( E . )     )      shift 9
E  -> E . op E    op     shift 1

i8:
E  -> E op E .           reduce 2
E  -> E . op E    op*    shift 1 *=if stack[-2] contains op of lower priority or if
                                   if it is of equal priority and right associative

i9:
E  -> ( E ) .            reduce 4

*/
/*}}}*/
static struct Value *eval(struct Value *value, const char *desc) /*{{{*/
{
  /* variables */ /*{{{*/
  static const int gotoState[10]={ 5,8,6,7,-1,-1,-1,-1,-1,-1 };
  int capacity=10;
  struct Pdastack
  {
    union
    {
      enum TokenType token;
      struct Value value;
    } u;
    char state;
  };
  struct Pdastack *pdastack=malloc(capacity*sizeof(struct Pdastack));
  struct Pdastack *sp=pdastack;
  struct Pdastack *send=pdastack+capacity-1;
  enum TokenType ip;
  /*}}}*/

  sp->state=0;
  while (1)
  {
    if (sp==send)
    {
      pdastack=realloc(pdastack,(capacity+10)*sizeof(struct Pdastack));
      sp=pdastack+capacity-1;
      capacity+=10;
      send=pdastack+capacity-1;
    }
    ip=pc.token->type;
    switch (sp->state)
    {
      case 0:
      case 1:
      case 2:
      case 3: /*{{{*/ /* including 4 */
      {
        if (ip==T_IDENTIFIER) /*{{{*/
        {
          /* printf("state %d: shift 4\n",sp->state); */
          /* printf("state 4: reduce E -> value\n"); */
          ++sp;
          sp->state=gotoState[(sp-1)->state];
          if (pass==COMPILE)
          {
            if
            (
              ((pc.token+1)->type==T_OP || Auto_find(&stack,pc.token->u.identifier)==0)
              && Global_find(&globals,pc.token->u.identifier,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR)==0
            )
            {
              Value_new_ERROR(value,UNDECLARED);
              goto error;
            }
          }
          if (pass!=DECLARE && (pc.token->u.identifier->sym->type==GLOBALVAR || pc.token->u.identifier->sym->type==LOCALVAR))
          {
            struct Value *l;

            if ((l=lvalue(value))->type==V_ERROR) goto error;
            Value_clone(&sp->u.value,l);
          }
          else
          {
            struct Pc var=pc;

            func(&sp->u.value);
            if (sp->u.value.type==V_VOID)
            {
              pc=var;
              Value_new_ERROR(value,VOIDVALUE);
              goto error;
            }
          }
        }
        /*}}}*/
        else if (ip==T_INTEGER) /*{{{*/
        {
          /* printf("state %d: shift 4\n",sp->state); */
          /* printf("state 4: reduce E -> value\n"); */
          ++sp;
          sp->state=gotoState[(sp-1)->state];
          VALUE_NEW_INTEGER(&sp->u.value,pc.token->u.integer);
          ++pc.token;
        }
        /*}}}*/
        else if (ip==T_REAL) /*{{{*/
        {
          /* printf("state %d: shift 4\n",sp->state); */
          /* printf("state 4: reduce E -> value\n"); */
          ++sp;
          sp->state=gotoState[(sp-1)->state];
          VALUE_NEW_REAL(&sp->u.value,pc.token->u.real);
          ++pc.token;
        }
        /*}}}*/
        else if (TOKEN_ISUNARYOPERATOR(ip)) /*{{{*/
        {
          /* printf("state %d: shift 2\n",sp->state); */
          ++sp;
          sp->state=2;
          sp->u.token=ip;
          ++pc.token;
        }
        /*}}}*/
        else if (ip==T_HEXINTEGER) /*{{{*/
        {
          /* printf("state %d: shift 4\n",sp->state); */
          /* printf("state 4: reduce E -> value\n"); */
          ++sp;
          sp->state=gotoState[(sp-1)->state];
          VALUE_NEW_INTEGER(&sp->u.value,pc.token->u.hexinteger);
          ++pc.token;
        }
        /*}}}*/
        else if (ip==T_OP) /*{{{*/
        {
          /* printf("state %d: shift 3\n",sp->state); */
          ++sp;
          sp->state=3;
          sp->u.token=T_OP;
          ++pc.token;
        }
        /*}}}*/
        else if (ip==T_STRING) /*{{{*/
        {
          /* printf("state %d: shift 4\n",sp->state); */
          /* printf("state 4: reduce E -> value\n"); */
          ++sp;
          sp->state=gotoState[(sp-1)->state];
          Value_new_STRING(&sp->u.value);
          String_destroy(&sp->u.value.u.string);
          String_clone(&sp->u.value.u.string,pc.token->u.string);
          ++pc.token;
        }
        /*}}}*/
        else /*{{{*/
        {
          char state=sp->state;

          if (state==0)
          {
            if (desc) Value_new_ERROR(value,MISSINGEXPR,desc);
            else value=(struct Value*)0;
          }
          else Value_new_ERROR(value,MISSINGEXPR,"operand");
          goto error;
        }
        /*}}}*/
        break;
      }
      /*}}}*/
      case 5: /*{{{*/
      {
        if (TOKEN_ISBINARYOPERATOR(ip))
        {
          /* printf("state %d: shift 1\n",sp->state); */
          ++sp;
          sp->state=1;
          sp->u.token=ip;
          ++pc.token;
          break;
        }
        else
        {
          assert(sp==pdastack+1);
          *value=sp->u.value;
          free(pdastack);
          return value;
        }
        break;
      }
      /*}}}*/
      case 6: /*{{{*/
      {
        if (TOKEN_ISBINARYOPERATOR(ip) && TOKEN_UNARYPRIORITY((sp-1)->u.token)<TOKEN_BINARYPRIORITY(ip))
        {
          assert(TOKEN_ISUNARYOPERATOR((sp-1)->u.token));
          /* printf("state %d: shift 1 (not reducing E -> op E)\n",sp->state); */
          ++sp;
          sp->state=1;
          sp->u.token=ip;
          ++pc.token;
        }
        else
        {
          enum TokenType op;

          /* printf("reduce E -> op E\n"); */
          --sp;
          op=sp->u.token;
          sp->u.value=(sp+1)->u.value;
          switch (op)
          {
            case T_PLUS: break;
            case T_MINUS: Value_uneg(&sp->u.value,pass==INTERPRET); break;
            case T_NOT: Value_unot(&sp->u.value,pass==INTERPRET); break;
            default: assert(0);
          }
          sp->state=gotoState[(sp-1)->state];
          if (sp->u.value.type==V_ERROR)
          {
            *value=sp->u.value;
            --sp;
            goto error;
          }
        }
        break;
      }
      /*}}}*/
      case 7: /*{{{*/ /* including 9 */
      {
        if (TOKEN_ISBINARYOPERATOR(ip))
        {
          /* printf("state %d: shift 1\n"sp->state); */
          ++sp;
          sp->state=1;
          sp->u.token=ip;
          ++pc.token;
        }
        else if (ip==T_CP)
        {
          /* printf("state %d: shift 9\n",sp->state); */
          /* printf("state 9: reduce E -> ( E )\n"); */
          --sp;
          sp->state=gotoState[(sp-1)->state];
          sp->u.value=(sp+1)->u.value;
          ++pc.token;
        }
        else
        {
          Value_new_ERROR(value,MISSINGCP);
          goto error;
        }
        break;
      }
      /*}}}*/
      case 8: /*{{{*/
      {
        int p1,p2;

        if
        (
          TOKEN_ISBINARYOPERATOR(ip)
          &&
          (
            ((p1=TOKEN_BINARYPRIORITY((sp-1)->u.token))<(p2=TOKEN_BINARYPRIORITY(ip)))
            || (p1==p2 && TOKEN_ISRIGHTASSOCIATIVE((sp-1)->u.token))
          )
        )
        {
          /* printf("state %d: shift 1\n",sp->state); */
          ++sp;
          sp->state=1;
          sp->u.token=ip;
          ++pc.token;
        }
        else
        {
          /* printf("state %d: reduce E -> E op E\n",sp->state); */
          if (Value_commonType[(sp-2)->u.value.type][sp->u.value.type]==V_ERROR)
          {
            Value_destroy(&sp->u.value);
            sp-=2;
            Value_destroy(&sp->u.value);
            Value_new_ERROR(value,INVALIDOPERAND);
            --sp;
            goto error;
          }
          else switch ((sp-1)->u.token)
          {
            case T_LT:    Value_lt(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_LE:    Value_le(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_EQ:    Value_eq(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_GE:    Value_ge(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_GT:    Value_gt(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_NE:    Value_ne(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_PLUS:  Value_add(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_MINUS: Value_sub(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_MULT:  Value_mult(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_DIV:   Value_div(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_IDIV:  Value_idiv(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_MOD:   Value_mod(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_POW:   Value_pow(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_AND:   Value_and(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_OR:    Value_or(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_XOR:   Value_xor(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_EQV:   Value_eqv(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            case T_IMP:   Value_imp(&(sp-2)->u.value,&sp->u.value,pass==INTERPRET); break;
            default:      assert(0);
          }
          Value_destroy(&sp->u.value);
          sp-=2;
          sp->state=gotoState[(sp-1)->state];
          if (sp->u.value.type==V_ERROR)
          {
            *value=sp->u.value;
            --sp;
            goto error;
          }
        }
        break;
      }
      /*}}}*/
    }
  }
  error:
  while (sp>pdastack)
  {
    switch (sp->state)
    {
      case 5:
      case 6:
      case 7:
      case 8: Value_destroy(&sp->u.value);
    }
    --sp;
  }
  free(pdastack);
  return value;
}
/*}}}*/
#else
static inline struct Value *binarydown(struct Value *value, struct Value *(level)(struct Value *value), const int prio) /*{{{*/
{
  enum TokenType op;
  struct Pc oppc;

  if (level(value)==(struct Value*)0) return (struct Value*)0;
  if (value->type==V_ERROR) return value;
  do
  {
    struct Value x;

    op=pc.token->type;
    if (!TOKEN_ISBINARYOPERATOR(op) || TOKEN_BINARYPRIORITY(op)!=prio) return value;
    oppc=pc;
    ++pc.token;
    if (level(&x)==(struct Value*)0)
    {
      Value_destroy(value);
      return Value_new_ERROR(value,MISSINGEXPR,"binary operand");
    }
    if (x.type==V_ERROR)
    {
      Value_destroy(value);
      *value=x;
      return value;
    }
    if (Value_commonType[value->type][x.type]==V_ERROR)
    {
      Value_destroy(value);
      Value_destroy(&x);
      return Value_new_ERROR(value,INVALIDOPERAND);
    }
    else switch (op)
    {
      case T_LT:    Value_lt(value,&x,pass==INTERPRET);   break;
      case T_LE:    Value_le(value,&x,pass==INTERPRET);   break;
      case T_EQ:    Value_eq(value,&x,pass==INTERPRET);   break;
      case T_GE:    Value_ge(value,&x,pass==INTERPRET);   break;
      case T_GT:    Value_gt(value,&x,pass==INTERPRET);   break;
      case T_NE:    Value_ne(value,&x,pass==INTERPRET);   break;
      case T_PLUS:  Value_add(value,&x,pass==INTERPRET);  break;
      case T_MINUS: Value_sub(value,&x,pass==INTERPRET);  break;
      case T_MULT:  Value_mult(value,&x,pass==INTERPRET); break;
      case T_DIV:   Value_div(value,&x,pass==INTERPRET);  break;
      case T_IDIV:  Value_idiv(value,&x,pass==INTERPRET); break;
      case T_MOD:   Value_mod(value,&x,pass==INTERPRET);  break;
      case T_POW:   Value_pow(value,&x,pass==INTERPRET);  break;
      case T_AND:   Value_and(value,&x,pass==INTERPRET);  break;
      case T_OR:    Value_or(value,&x,pass==INTERPRET);   break;
      case T_XOR:   Value_xor(value,&x,pass==INTERPRET);  break;
      case T_EQV:   Value_eqv(value,&x,pass==INTERPRET);  break;
      case T_IMP:   Value_imp(value,&x,pass==INTERPRET);  break;
      default:      assert(0);
    }
    Value_destroy(&x);
  } while (value->type!=V_ERROR);
  if (value->type==V_ERROR) pc=oppc;
  return value;
}
/*}}}*/
static inline struct Value *unarydown(struct Value *value, struct Value *(level)(struct Value *value), const int prio) /*{{{*/
{
  enum TokenType op;
  struct Pc oppc;

  op=pc.token->type;
  if (!TOKEN_ISUNARYOPERATOR(op) || TOKEN_UNARYPRIORITY(op)!=prio) return level(value);
  oppc=pc;
  ++pc.token;
  if (unarydown(value,level,prio)==(struct Value*)0) return Value_new_ERROR(value,MISSINGEXPR,"unary operand");
  if (value->type==V_ERROR) return value;
  switch (op)
  {
    case T_PLUS:  Value_uplus(value,pass==INTERPRET); break;
    case T_MINUS: Value_uneg(value,pass==INTERPRET);  break;
    case T_NOT:   Value_unot(value,pass==INTERPRET);  break;
    default:      assert(0);
  }
  if (value->type==V_ERROR) pc=oppc;
  return value;
}
/*}}}*/
static struct Value *eval8(struct Value *value) /*{{{*/
{
  switch (pc.token->type)
  {
    case T_IDENTIFIER:
    {
      struct Pc var;
      struct Value *l;

      var=pc;
      if (pass==COMPILE)
      {
        if
        (
          ((pc.token+1)->type==T_OP || Auto_find(&stack,pc.token->u.identifier)==0)
          && Global_find(&globals,pc.token->u.identifier,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR)==0
        ) return Value_new_ERROR(value,UNDECLARED);
      }
      if (pass!=DECLARE && (pc.token->u.identifier->sym->type==GLOBALVAR || pc.token->u.identifier->sym->type==LOCALVAR))
      {
        if ((l=lvalue(value))->type==V_ERROR) return value;
        Value_clone(value,l);
      }
      else
      {
        func(value);
        if (value->type==V_VOID)
        {
          Value_destroy(value);
          pc=var;
          return Value_new_ERROR(value,VOIDVALUE);
        }
      }
      break;
    }
    case T_INTEGER:
    {
      VALUE_NEW_INTEGER(value,pc.token->u.integer);
      ++pc.token;
      break;
    }
    case T_REAL:
    {
      VALUE_NEW_REAL(value,pc.token->u.real);
      ++pc.token;
      break;
    }
    case T_STRING:
    {
      Value_new_STRING(value);
      String_destroy(&value->u.string);
      String_clone(&value->u.string,pc.token->u.string);
      ++pc.token;
      break;
    }
    case T_HEXINTEGER:
    {
      VALUE_NEW_INTEGER(value,pc.token->u.hexinteger);
      ++pc.token;
      break;
    }
    case T_OP:
    {
      ++pc.token;
      if (eval(value,"parenthetic")->type==V_ERROR) return value;
      if (pc.token->type!=T_CP)
      {
        Value_destroy(value);
        return Value_new_ERROR(value,MISSINGCP);
      }
      ++pc.token;
      break;
    }
    default:
    {
      return (struct Value *)0;
    }
  }
  return value;
}
/*}}}*/
static struct Value *eval7(struct Value *value) /*{{{*/
{
  return binarydown(value,eval8,7);
}
/*}}}*/
static struct Value *eval6(struct Value *value) /*{{{*/
{
  return unarydown(value,eval7,6);
}
/*}}}*/
static struct Value *eval5(struct Value *value) /*{{{*/
{
  return binarydown(value,eval6,5);
}
/*}}}*/
static struct Value *eval4(struct Value *value) /*{{{*/
{
  return binarydown(value,eval5,4);
}
/*}}}*/
static struct Value *eval3(struct Value *value) /*{{{*/
{
  return binarydown(value,eval4,3);
}
/*}}}*/
static struct Value *eval2(struct Value *value) /*{{{*/
{
  return unarydown(value,eval3,2);
}
/*}}}*/
static struct Value *eval1(struct Value *value) /*{{{*/
{
  return binarydown(value,eval2,1);
}
/*}}}*/
static struct Value *eval(struct Value *value, const char *desc) /*{{{*/
{
  /* avoid function calls for atomic expression */
  switch (pc.token->type)
  {
    case T_STRING:
    case T_REAL:
    case T_INTEGER:
    case T_HEXINTEGER:
    case T_IDENTIFIER: if (!TOKEN_ISBINARYOPERATOR((pc.token+1)->type) && (pc.token+1)->type!=T_OP) return eval7(value);
  }
  if (binarydown(value,eval1,0)==(struct Value*)0)
  {
    if (desc) return Value_new_ERROR(value,MISSINGEXPR,desc);
    else return (struct Value*)0; 
  }
  else return value;
}
/*}}}*/
#endif

static void new(void) /*{{{*/
{
  Global_destroy(&globals);
  Global_new(&globals);
  Program_destroy(&program);
  Program_new(&program);
  FS_closefiles();
  runnable=0;
}
/*}}}*/
static void dobreak(int sig) /*{{{*/
{
  stop=1;
  allowBreak(0);
}
/*}}}*/
static void allowBreak(int on) /*{{{*/
{
  struct sigaction breakact;

  breakact.sa_handler=on ? dobreak : SIG_IGN;
  sigemptyset(&breakact.sa_mask);
  sigaddset(&breakact.sa_mask,SIGINT);
  breakact.sa_flags=0;
  sigaction(SIGINT,&breakact,(struct sigaction *)0);
}
/*}}}*/
static void pushLabel(enum LabelType type, struct Pc *patch) /*{{{*/
{
  if (labelStackPointer==labelStackCapacity)
  {
    labelStack=realloc(labelStack,sizeof(struct LabelStack)*(labelStackCapacity?(labelStackCapacity*=2):(32)));
  }
  labelStack[labelStackPointer].type=type;
  labelStack[labelStackPointer].patch=*patch;
  ++labelStackPointer;
}
/*}}}*/
static struct Pc *popLabel(enum LabelType type) /*{{{*/
{
  if (labelStackPointer==0 || labelStack[labelStackPointer-1].type!=type) return (struct Pc*)0;
  else return &labelStack[--labelStackPointer].patch;
}
/*}}}*/
static struct Pc *findLabel(enum LabelType type) /*{{{*/
{
  unsigned int i;

  for (i=0; i<labelStackPointer; ++i) if (labelStack[i].type==type) return &labelStack[i].patch;
  return (struct Pc*)0;
}
/*}}}*/
static void labelStackError(struct Value *v) /*{{{*/
{
  struct String s;

  assert(labelStackPointer);
  pc=labelStack[labelStackPointer-1].patch;
  switch (labelStack[labelStackPointer-1].type)
  {
    case L_IF: Value_new_ERROR(v,STRAYIF); break;
    case L_DO: Value_new_ERROR(v,STRAYDO); break;
    case L_DOWHILE: Value_new_ERROR(v,STRAYDOWHILE); break;
    case L_ELSE: Value_new_ERROR(v,STRAYELSE2); break;
    case L_FOR_BODY:
    {
      Value_new_ERROR(v,STRAYFOR);
      pc=*findLabel(L_FOR);
      break;
    }
    case L_WHILE: Value_new_ERROR(v,STRAYWHILE); break;
    case L_REPEAT: Value_new_ERROR(v,STRAYREPEAT); break;
    case L_SELECTCASE: Value_new_ERROR(v,STRAYSELECTCASE); break;
    default: assert(0);
  }
}
/*}}}*/
static struct Value *assign(struct Value *value) /*{{{*/
{
  struct Value *l;
  struct Pc var,expr;

  var=pc;
  if (pass==DECLARE)
  {
    if
    (
      ((pc.token+1)->type==T_OP || Auto_find(&stack,pc.token->u.identifier)==0)
      && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR,0)==0
    )
    {
      return Value_new_ERROR(value,REDECLARATION);
    }
  }
  if ((l=lvalue(value))->type==V_ERROR) return value;
  if (pc.token->type!=T_EQ) return Value_new_ERROR(value,MISSINGEQ);
  ++pc.token;
  expr=pc;
  if (eval(value,"rhs")->type==V_ERROR) return value;
  if (pass!=DECLARE && Value_retype(value,l->type)->type==V_ERROR)
  {
    pc=expr;
    return value;
  }
  if (pass==INTERPRET)
  {
    Value_destroy(l);
    *l=*value;
  }
  return value;
}
/*}}}*/
static struct Value *compile(struct Value *v) /*{{{*/
{
  struct Pc begin;

  stack.resumeable=0;
  if (Program_beginning(&program,&begin)==(struct Pc*)0)
  {
    return Value_new_ERROR(v,NOPROGRAM);
  }
  else
  {
    struct Value value;
    struct Pc savepc;
    int savepass;

    savepc=pc;
    savepass=pass;
    Global_destroy(&globals);
    Global_new(&globals);
    for (pass=DECLARE; pass!=INTERPRET; ++pass)
    {
      if (pass==DECLARE)
      {
        begindata.line=-1;
        lastdata=&begindata;
      }
      stop=0;
      runnable=1;
      pc=begin;
      while (1)
      {
        statements(v);
        if (v->type==V_ERROR) break;
        Value_destroy(v);
        if (!Program_skipEOL(&program,&pc,0,0)) { Value_new_NIL(v); break; }
      }
      if (v->type!=V_ERROR && labelStackPointer>0)
      {
        Value_destroy(v);
        labelStackError(v);
      }
      if (v->type==V_ERROR)
      {
        labelStackPointer=0;
        runnable=0;
        pass=savepass;
        return v;
      }
    }
    curdata=begindata;
    pc=savepc;
    pass=savepass;
    return Value_new_NIL(v);
  }
}
/*}}}*/
static void runline(struct Token *line) /*{{{*/
{
  struct Value value;

  FS_flush(STDCHANNEL);
  for (pass=DECLARE; pass!=INTERPRET; ++pass)
  {
    pc.line=-1;
    pc.token=line;
    stop=0;
    statements(&value);
    if (value.type!=V_ERROR && pc.token->type!=T_EOL)
    {
      Value_destroy(&value);
      Value_new_ERROR(&value,SYNTAX);
    }
    if (value.type!=V_ERROR && labelStackPointer>0)
    {
      Value_destroy(&value);
      labelStackError(&value);
    }
    if (value.type==V_ERROR)
    {
      struct String s;

      Auto_setError(&stack,Program_lineNumber(&program,&pc),&pc,&value);
      Program_PCtoError(&program,&pc,&value);
      labelStackPointer=0;
      FS_putChars(STDCHANNEL,"error: ");
      String_new(&s);
      Value_toString(&value,&s,' ',-1,0,0,-1,0);
      Value_destroy(&value);
      FS_putString(STDCHANNEL,&s);
      String_destroy(&s);
      return;
    }
  }

  pc.line=-1;
  pc.token=line;
  curdata=begindata;
  nextdata.line=-1;
  Value_destroy(&value);
  pass=INTERPRET;
  allowBreak(1);
  do
  {
    assert(pass==INTERPRET);
    statements(&value);
    assert(pass==INTERPRET);
    if (value.type==V_ERROR)
    {
      if (strchr(value.u.error.msg,'\n')==(char*)0)
      {
        Auto_setError(&stack,Program_lineNumber(&program,&pc),&pc,&value);
        Program_PCtoError(&program,&pc,&value);
      }
      if (stack.onerror.line!=-1)
      {
        stack.resumeable=1;
        pc=stack.onerror;
      }
      else
      {
        struct String s;

        String_new(&s);
        FS_putChars(STDCHANNEL,"error: ");
        Auto_frameToError(&stack,&program,&value);
        Value_toString(&value,&s,' ',-1,0,0,-1,0);
        while (Auto_gosubReturn(&stack,(struct Pc*)0));
        FS_putString(STDCHANNEL,&s);
        String_destroy(&s);
        Value_destroy(&value);
        break;
      }
    }
    Value_destroy(&value);
  } while (pc.token->type!=T_EOL || Program_skipEOL(&program,&pc,STDCHANNEL,1));
  allowBreak(0);
}
/*}}}*/
static struct Value *evalGeometry(struct Value *value, unsigned int geometry[]) /*{{{*/
{
  struct Pc exprpc=pc;

  if (eval(value,"dimension")->type==V_ERROR || (pass!=DECLARE && Value_retype(value,V_INTEGER)->type==V_ERROR)) return value;
  if (pass==INTERPRET && value->u.integer<0)
  {
    Value_destroy(value);
    pc=exprpc;
    return Value_new_ERROR(value,OUTOFRANGE,"dimension");
  }
  geometry[0]=value->u.integer+1;
  Value_destroy(value);
  if (pc.token->type==T_COMMA)
  {
    ++pc.token;
    exprpc=pc;
    if (eval(value,"dimension")->type==V_ERROR || (pass!=DECLARE && Value_retype(value,V_INTEGER)->type==V_ERROR)) return value;
    if (pass==INTERPRET && value->u.integer<0)
    {
      Value_destroy(value);
      pc=exprpc;
      return Value_new_ERROR(value,OUTOFRANGE,"dimension");
    }
    geometry[1]=value->u.integer+1;
    Value_destroy(value);
  }
  else geometry[1]=2;
  return (struct Value*)0;
}
/*}}}*/
static struct Value *convert(struct Value *value, struct Value *l, struct Token *t) /*{{{*/
{
  switch (l->type)
  {
    case V_INTEGER:
    {
      char *datainput;
      char *end;
      long int v;

      if (t->type!=T_DATAINPUT) return Value_new_ERROR(value,BADCONVERSION,"integer");
      datainput=t->u.datainput;
      if (datainput[0]=='&' && (datainput[1]=='h' || datainput[1]=='H')) v=strtoul(datainput+=2,&end,16);
      else v=strtol(datainput,&end,10);
      if (end==datainput || (*end!='\0' && *end!=' ' && *end!='\t')) return Value_new_ERROR(value,BADCONVERSION,"integer");
      Value_destroy(l);
      VALUE_NEW_INTEGER(l,v);
      break;
    }
    case V_REAL:
    {
      char *datainput;
      char *end;
      double v;

      if (t->type!=T_DATAINPUT) return Value_new_ERROR(value,BADCONVERSION,"real");
      datainput=t->u.datainput;
      v=strtod(datainput,&end);
      if (end==datainput || (*end!='\0' && *end!=' ' && *end!='\t')) return Value_new_ERROR(value,BADCONVERSION,"real");
      Value_destroy(l);
      VALUE_NEW_REAL(l,v);
      break;
    }
    case V_STRING:
    {
      Value_destroy(l);
      Value_new_STRING(l);
      if (t->type==T_STRING) String_appendString(&l->u.string,t->u.string);
      else String_appendChars(&l->u.string,t->u.datainput);
      break;
    }
    default: assert(0);
  }
  return (struct Value*)0;
}
/*}}}*/
static struct Value *dataread(struct Value *value, struct Value *l) /*{{{*/
{
  if (curdata.line==-1)
  {
    return Value_new_ERROR(value,ENDOFDATA);
  }
  if (curdata.token->type==T_DATA)
  {
    nextdata=curdata.token->u.nextdata;
    ++curdata.token;
  }
  if (convert(value,l,curdata.token))
  {
    return value;
  }
  ++curdata.token;
  if (curdata.token->type==T_COMMA) ++curdata.token;
  else curdata=nextdata;
  return (struct Value*)0;
}
/*}}}*/
static struct Value *statements(struct Value *value) /*{{{*/
{
  struct Pc statementpc=pc;

  more:
  switch (pc.token->type)
  {
    case T_CALL: /*{{{*/
    {
      struct Pc here;
      const char *name;

      ++pc.token;
      if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"procedure");
      here=pc;
      name=pc.token->u.identifier->name;
      if (pass==DECLARE)
      {
        if (func(value)->type==V_ERROR) return value;
        else Value_destroy(value);
      }
      else
      {
        if (pass==COMPILE)
        {
          if
          (
            Global_find(&globals,pc.token->u.identifier,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR)==0
          ) return Value_new_ERROR(value,UNDECLARED);
        }
        if (pc.token->u.identifier->sym->type!=USERFUNCTION) return Value_new_ERROR(value,TYPEMISMATCH1,"variable","function");
        func(value);
        if (Value_retype(value,V_VOID)->type==V_ERROR) return value;
        Value_destroy(value);
      }
      break;
    }
    /*}}}*/
    case T_CASEELSE: 
    case T_CASEVALUE: /*{{{*/
    {
      if (pass==DECLARE || pass==COMPILE)
      {
        struct Pc *selectcase,*nextcasevalue;

        if ((selectcase=findLabel(L_SELECTCASE))==(struct Pc*)0) return Value_new_ERROR(value,STRAYCASE);
        for (nextcasevalue=&selectcase->token->u.selectcase->nextcasevalue; nextcasevalue->line!=-1; nextcasevalue=&nextcasevalue->token->u.casevalue->nextcasevalue);
        *nextcasevalue=pc;
        if (pass==COMPILE) pc.token->u.casevalue->endselect=selectcase->token->u.selectcase->endselect;
        pc.token->u.casevalue->nextcasevalue.line=-1;
        ++pc.token; 
        switch (statementpc.token->type)
        {
          case T_CASEELSE: break;
          case T_CASEVALUE:
          {
            struct Pc exprpc;

            do
            {
              exprpc=pc;
              if (eval(value,"case")->type==V_ERROR) return value;
              if (Value_retype(value,selectcase->token->u.selectcase->type)->type==V_ERROR)
              {
                pc=exprpc;
                return value;
              }
              Value_destroy(value);
              if (pc.token->type==T_TO)
              {
                ++pc.token;
                exprpc=pc;
                if (eval(value,"case")->type==V_ERROR) return value;
                if (Value_retype(value,selectcase->token->u.selectcase->type)->type==V_ERROR)
                {
                  pc=exprpc;
                  return value;
                }
                Value_destroy(value);
              }
              if (pc.token->type==T_COMMA) ++pc.token;
              else break;
            } while (1);
            break;
          }
          default: assert(0);
        }
      }
      else pc=pc.token->u.casevalue->endselect;
      break;
    }
    /*}}}*/
    case T_CHDIR:
    case T_MKDIR: /*{{{*/
    {
      int res,err;
      struct Pc dirpc;

      ++pc.token;
      dirpc=pc;
      if (eval(value,"directory")->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR) return value;
      if (pass==INTERPRET)
      {
        switch (statementpc.token->type)
        {
          case T_CHDIR: res=chdir(value->u.string.character); break;
          case T_MKDIR: res=mkdir(value->u.string.character); break;
          default: assert(0);
        }
        err=errno;
      }
      Value_destroy(value);
      if (pass==INTERPRET && res==-1)
      {
        pc=dirpc;
        return Value_new_ERROR(value,IOERROR,strerror(err));
      }
      break;
    }
    /*}}}*/
    case T_CLEAR: /*{{{*/
    {
      if (pass==INTERPRET) Global_clear(&globals);
      ++pc.token;
      break;
    }
    /*}}}*/
    case T_CLOSE: /*{{{*/
    {
      int hasargs=0;
      struct Pc chnpc;

      ++pc.token;
      while (1)
      {
        chnpc=pc;
        if (pc.token->type==T_CHANNEL) { hasargs=1; ++pc.token; }
        if (eval(value,(const char*)0)==(struct Value*)0)
        {
          if (hasargs) return Value_new_ERROR(value,MISSINGEXPR,"channel");
          else break;
        }
        hasargs=1;
        if (value->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        if (pass==INTERPRET && FS_close(value->u.integer)==-1)
        {
          Value_destroy(value);
          pc=chnpc;
          return Value_new_ERROR(value,IOERROR,FS_errmsg);
        }
        if (pc.token->type==T_COMMA) ++pc.token;
        else break;
      }
      if (!hasargs && pass==INTERPRET) FS_closefiles();
      break;
    }
    /*}}}*/
    case T_CLS: /*{{{*/
    {
      ++pc.token;
      if (pass==INTERPRET && FS_cls(STDCHANNEL)==-1)
      {
        pc=statementpc;
        return Value_new_ERROR(value,IOERROR,FS_errmsg);
      }
      break;
    }
    /*}}}*/
    case T_DATA: /*{{{*/
    {
      if (DIRECTMODE) return Value_new_ERROR(value,NOTINDIRECTMODE);
      if (pass==DECLARE)
      {
        *lastdata=pc;
        (lastdata=&(pc.token->u.nextdata))->line=-1;
      }
      ++pc.token;
      while (1)
      {
        if (pc.token->type!=T_STRING && pc.token->type!=T_DATAINPUT) return Value_new_ERROR(value,MISSINGDATAINPUT);
        ++pc.token;
        if (pc.token->type!=T_COMMA) break;
        else ++pc.token;
      }
      break;
    }
    /*}}}*/
    case T_DEFFN:
    case T_DEFPROC:
    case T_FUNCTION:
    case T_SUB: /*{{{*/
    {
      if (pass==DECLARE || pass==COMPILE)
      {
        struct Identifier *fn;
        int proc;
        int args=0;

        if (DIRECTMODE) return Value_new_ERROR(value,NOTINDIRECTMODE);
        proc=(pc.token->type==T_DEFPROC || pc.token->type==T_SUB);
        ++pc.token;
        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,proc?"procedure":"function");
        fn=pc.token->u.identifier;
        if (proc) fn->defaultType=V_VOID;
        ++pc.token;
        if (findLabel(L_FUNC))
        {
          pc=statementpc;
          return Value_new_ERROR(value,NESTEDDEFINITION);
        }
        Auto_variable(&stack,fn);
        if (pc.token->type==T_OP) /* arguments */ /*{{{*/
        {
          ++pc.token;
          while (1)
          {
            if (pc.token->type!=T_IDENTIFIER)
            {
              Auto_funcEnd(&stack);
              return Value_new_ERROR(value,MISSINGIDENTIFIER,"formal parameter");
            }
            if (Auto_variable(&stack,pc.token->u.identifier)==0)
            {
              Auto_funcEnd(&stack);
              return Value_new_ERROR(value,ALREADYDECLARED);
            }
            ++args;
            ++pc.token;
            if (pc.token->type==T_COMMA) ++pc.token;
            else break;
          }
          if (pc.token->type!=T_CP)
          {
            Auto_funcEnd(&stack);
            return Value_new_ERROR(value,MISSINGCP);
          }
          ++pc.token;
        }
        /*}}}*/
        if (pass==DECLARE)
        {
          enum ValueType *t=args ? malloc(args*sizeof(enum ValueType)) : (enum ValueType*)0;
          int i;

          for (i=0; i<args; ++i) t[i]=Auto_argType(&stack,args-i);
          if (Global_function(&globals,fn,fn->defaultType,&pc,&statementpc,args,t)==0)
          {
            free(t);
            Auto_funcEnd(&stack);
            pc=statementpc;
            return Value_new_ERROR(value,REDECLARATION);
          }
          Program_addScope(&program,&fn->sym->u.sub.u.def.scope);
        }
        pushLabel(L_FUNC,&statementpc);
        if (pc.token->type==T_EQ) goto funcreturn;
      }
      else pc=(pc.token+1)->u.identifier->sym->u.sub.u.def.scope.end;
      break;
    }
    /*}}}*/
    case T_DEFINT:
    case T_DEFDBL:
    case T_DEFSTR: /*{{{*/
    {
      enum ValueType dsttype=V_NIL;

      switch (pc.token->type)
      {
        case T_DEFINT: dsttype=V_INTEGER; break;
        case T_DEFDBL: dsttype=V_REAL; break;
        case T_DEFSTR: dsttype=V_STRING; break;
        default: assert(0);
      }
      ++pc.token;
      while (1)
      {
        struct Identifier *ident;

        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"variable");
        if (pc.token->u.identifier->defaultType!=V_REAL) switch (dsttype)
        {
          case V_INTEGER: return Value_new_ERROR(value,BADIDENTIFIER,"integer");
          case V_REAL: return Value_new_ERROR(value,BADIDENTIFIER,"real");
          case V_STRING: return Value_new_ERROR(value,BADIDENTIFIER,"string");
          default: assert(0);
        }
        ident=pc.token->u.identifier;
        ++pc.token;
        if (pc.token->type==T_MINUS)
        {
          struct Identifier i;

          if (strlen(ident->name)!=1) return Value_new_ERROR(value,BADRANGE);
          ++pc.token;
          if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"variable");
          if (strlen(pc.token->u.identifier->name)!=1) return Value_new_ERROR(value,BADRANGE);
          for (i.name[0]=tolower(ident->name[0]),i.name[1]='\0'; i.name[0]<=tolower(pc.token->u.identifier->name[0]); ++i.name[0])
          {
            Global_variable(&globals,&i,dsttype,G_SCALAR,1);
          }
          ++pc.token;
        }
        else Global_variable(&globals,ident,dsttype,G_SCALAR,1);
        if (pc.token->type==T_COMMA) ++pc.token;
        else break;
      }
      break;
    }
    /*}}}*/
    case T_DELETE: /*{{{*/
    {
      if (pass==INTERPRET && !DIRECTMODE) return Value_new_ERROR(value,NOTINPROGRAMMODE);
      ++pc.token;
      if (pc.token->type==T_INTEGER)
      {
        if (pass==INTERPRET)
        {
          if (!program.numbered) return Value_new_ERROR(value,UNNUMBERED);
          else if (pc.token->u.integer<=0) return Value_new_ERROR(value,OUTOFRANGE,"line number");
          else if (Program_delete(&program,pc.token->u.integer)==-1) return Value_new_ERROR(value,NOSUCHLINE);
        }
        ++pc.token;
      }
      break;
    }
    /*}}}*/
    case T_DIM: /*{{{*/
    {
      ++pc.token;
      while (1)
      {
        unsigned int capacity=0,*geometry=(unsigned int*)0;
        struct Var *var;
        struct Pc dimpc;
        unsigned int dim;
        enum ValueType vartype;

        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"array variable");
        if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,G_ARRAY,0)==0)
        {
          return Value_new_ERROR(value,REDECLARATION);
        }
        var=&pc.token->u.identifier->sym->u.var;
        if (pass==INTERPRET && var->dim) return Value_new_ERROR(value,REDIM);
        vartype=var->type;
        ++pc.token;
        if (pc.token->type!=T_OP) return Value_new_ERROR(value,MISSINGOP);
        ++pc.token;
        dim=0;
        while (1)
        {
          dimpc=pc;
          if (eval(value,"dimension")->type==V_ERROR || (pass!=DECLARE && Value_retype(value,V_INTEGER)->type==V_ERROR))
          {
            if (capacity) free(geometry);
            return value;
          }
          if (pass==INTERPRET && value->u.integer<0) /* error */ /*{{{*/
          {
            Value_destroy(value);
            Value_new_ERROR(value,OUTOFRANGE,"dimension");
          }
          /*}}}*/
          if (value->type==V_ERROR) /* abort */ /*{{{*/
          {
            if (capacity) free(geometry);
            pc=dimpc;
            return value;
          }
          /*}}}*/
          if (pass==INTERPRET)
          {
            if (dim==capacity) /* enlarge geometry */ /*{{{*/
            {
              geometry=realloc(geometry,sizeof(unsigned int)*(capacity?(capacity*=2):(capacity=3)));
            }
            /*}}}*/
            geometry[dim]=value->u.integer+1;
            ++dim;
          }
          if (pc.token->type==T_COMMA)
          {
            ++pc.token;
            Value_destroy(value);
          }
          else break;
        }
        if (pc.token->type!=T_CP) /* abort */ /*{{{*/
        {
          if (capacity) free(geometry);
          Value_destroy(value);
          return Value_new_ERROR(value,MISSINGCP);
        }
        /*}}}*/
        ++pc.token;
        if (pass==INTERPRET)
        {
          Var_destroy(var);
          Var_new(var,vartype,dim,geometry);
          free(geometry);
        }
        if (pc.token->type==T_COMMA) /* prepare for next var */ /*{{{*/
        {
          ++pc.token;
          Value_destroy(value);
        }
        /*}}}*/
        else break;
      }
      break;
    }
    /*}}}*/
    case T_DO: /*{{{*/
    {
      if (pass==DECLARE || pass==COMPILE) pushLabel(L_DO,&pc);
      ++pc.token;
      break;
    }
    /*}}}*/
    case T_DOWHILE: /*{{{*/
    {
      struct Pc dowhilepc=pc;

      if (pass==DECLARE || pass==COMPILE) pushLabel(L_DOWHILE,&pc);
      ++pc.token;
      if (eval(value,"condition")->type==V_ERROR) return value;
      if (pass==INTERPRET)
      {
        if (Value_isNull(value)) pc=dowhilepc.token->u.exitdo;
        Value_destroy(value);
      }
      break;
    }
    /*}}}*/
    case T_EDIT: /*{{{*/
    {
      long int line;

      ++pc.token;
      if (pc.token->type==T_INTEGER)
      {
        struct Pc where;

        if (program.numbered)
        {
          if (Program_goLine(&program,pc.token->u.integer,&where)==(struct Pc*)0) return Value_new_ERROR(value,NOSUCHLINE);
          line=where.line+1;
        }
        else
        {
          Program_end(&program,&where);
          line=pc.token->u.integer;
          if (line<1 || line>(where.line+1)) return Value_new_ERROR(value,NOSUCHLINE);
        }
        ++pc.token;
      }
      else line=1;
      if (pass==INTERPRET)
      {
        /* variables */ /*{{{*/
        char *name;
        int chn;
        struct Program newProgram;
        const char *visual,*basename,*shell;
        struct String cmd;
        static struct
        {
          const char *editor,*flag;
        }
        gotoLine[]=
        {
          { "Xemacs", "+%ld " },
          { "cemacs", "+%ld " },
          { "emacs", "+%ld " },
          { "emori", "-l%ld " },
          { "fe", "-l%ld " },
          { "jed", "+%ld " },
          { "jmacs", "+%ld " },
          { "joe", "+%ld " },
          { "modeori", "-l%ld " },
          { "origami", "-l%ld " },
          { "vi", "-c%ld " }, 
          { "vim", "+%ld " },
          { "xemacs", "+%ld " }
        };
        unsigned int i;
        /*}}}*/

        if (!DIRECTMODE)
        {
          pc=statementpc;
          return Value_new_ERROR(value,NOTINPROGRAMMODE);
        }
        if ((name=tmpnam((char*)0))==(char*)0)
        {
          pc=statementpc;
          return Value_new_ERROR(value,IOERROR,"unable to generate temporary file name");
        }
        if ((chn=FS_openout(name,1))==-1)
        {
          pc=statementpc;
          return Value_new_ERROR(value,IOERROR2,"create",name,FS_errmsg);
        }
        if (Program_list(&program,chn,(struct Pc*)0,(struct Pc*)0,value))
        {
          pc=statementpc;
          return value;
        }
        if (FS_close(chn)==-1)
        {
          pc=statementpc;
          unlink(name);
          return Value_new_ERROR(value,IOERROR,"close",name,FS_errmsg);
        }
        if ((visual=getenv("VISUAL"))==(char*)0 && (visual=getenv("EDITOR"))==(char*)0) visual="vi";
        basename=strrchr(visual,'/');
        if (basename==(char*)0) basename=visual;
        if ((shell=getenv("SHELL"))==(char*)0) shell="/bin/sh";
        String_new(&cmd);
        String_appendChars(&cmd,visual);
        String_appendChar(&cmd,' ');
        for (i=0; i<sizeof(gotoLine)/sizeof(gotoLine[0]); ++i)
        {
          if (strcmp(basename,gotoLine[i].editor)==0)
          {
            String_appendPrintf(&cmd,gotoLine[i].flag,line);
            break;
          }
        }
        String_appendChars(&cmd,name);
        FS_shellmode(STDCHANNEL);
        switch (fork())
        {
          case -1:
          {
            unlink(name);
            FS_fsmode(STDCHANNEL);
            return Value_new_ERROR(value,FORKFAILED,strerror(errno));
          }
          case 0:
          {
            execl(shell,shell,"-c",cmd.character,(const char*)0);
            exit(127);
          }
          default:
          {
            wait((int*)0);
          }
        }
        FS_fsmode(STDCHANNEL);
        String_destroy(&cmd);
        if ((chn=FS_openin(name))==-1)
        {
          pc=statementpc;
          return Value_new_ERROR(value,IOERROR2,"open",name,FS_errmsg);
        }
        Program_new(&newProgram);
        if (Program_merge(&newProgram,chn,value))
        {
          FS_close(chn);
          unlink(name);
          pc=statementpc;
          return value;
        }
        FS_close(chn);
        Program_destroy(&program);
        program=newProgram;
        runnable=0;
        unlink(name);
      }
      break;
    }
    /*}}}*/
    case T_ELSE:
    case T_ELSEIFELSE: /*{{{*/
    {
      if (pass==INTERPRET)
      {
        pc=pc.token->u.endifpc;
      }
      if (pass==DECLARE || pass==COMPILE)
      {
        struct Pc elsepc=pc;
        struct Pc *ifinstr;
        int elseifelse=(pc.token->type==T_ELSEIFELSE);

        if ((ifinstr=popLabel(L_IF))==(struct Pc*)0) return Value_new_ERROR(value,STRAYELSE1);
        if (ifinstr->token->type==T_ELSEIFIF) (ifinstr->token-1)->u.elsepc=pc;
        ++pc.token;
        ifinstr->token->u.elsepc=pc;
        assert(ifinstr->token->type==T_ELSEIFIF || ifinstr->token->type==T_IF);
        if (elseifelse) goto more;
        else pushLabel(L_ELSE,&elsepc);
      }
      break;
    }
    /*}}}*/
    case T_END: /*{{{*/
    {
      if (pass==INTERPRET) pc=pc.token->u.endpc;
      if (pass==DECLARE || pass==COMPILE)
      {
        Program_end(&program,&pc.token->u.endpc);
        ++pc.token;
      }
      break;
    }
    /*}}}*/
    case T_ENDIF: /*{{{*/
    {
      if (pass==DECLARE || pass==COMPILE)
      {
        struct Pc endifpc=pc;
        struct Pc *ifpc;
        struct Pc *elsepc;

        if ((ifpc=popLabel(L_IF)))
        {
          ifpc->token->u.elsepc=endifpc;
          if (ifpc->token->type==T_ELSEIFIF) (ifpc->token-1)->u.elsepc=pc;
        }
        else if ((elsepc=popLabel(L_ELSE))) elsepc->token->u.endifpc=endifpc;
        else return Value_new_ERROR(value,STRAYENDIF);
      }
      ++pc.token;
      break;
    }
    /*}}}*/
    case T_ENDFN: /*{{{*/
    {
      struct Pc *curfn=(struct Pc*)0;
      struct Pc eqpc=pc;

      if (pass==DECLARE || pass==COMPILE)
      {
        if ((curfn=popLabel(L_FUNC))==(struct Pc*)0) return Value_new_ERROR(value,STRAYENDFN);
        if ((eqpc.token->u.type=(curfn->token+1)->u.identifier->defaultType)==V_VOID) return Value_new_ERROR(value,STRAYENDFN);
      }
      ++pc.token;
      if (pass==INTERPRET) return Value_clone(value,Var_value(Auto_local(&stack,0),0,(unsigned int*)0,(struct Value*)0));
      else
      {
        if (pass==DECLARE) Global_endfunction(&globals,(curfn->token+1)->u.identifier,&pc);
        Auto_funcEnd(&stack);
      }
      break;
    }
    /*}}}*/
    case T_ENDPROC:
    case T_ENDSUB: /*{{{*/
    {
      struct Pc *curfn=(struct Pc*)0;

      if (pass==DECLARE || pass==COMPILE)
      {
        if ((curfn=popLabel(L_FUNC))==(struct Pc*)0 || (curfn->token+1)->u.identifier->defaultType!=V_VOID)
        {
          if (pc.token->type==T_ENDPROC) return Value_new_ERROR(value,STRAYENDPROC);
          else return Value_new_ERROR(value,STRAYENDSUB);
        }
      }
      ++pc.token;
      if (pass==INTERPRET) return Value_new_VOID(value);
      else
      {
        if (pass==DECLARE) Global_endfunction(&globals,(curfn->token+1)->u.identifier,&pc);
        Auto_funcEnd(&stack);
      }
      break;
    }
    /*}}}*/
    case T_ENDSELECT: /*{{{*/
    {
      ++pc.token;
      if (pass==DECLARE || pass==COMPILE)
      {
        struct Pc *selectcasepc;

        if ((selectcasepc=popLabel(L_SELECTCASE))) selectcasepc->token->u.selectcase->endselect=pc;
        else
        {
          pc=statementpc;
          return Value_new_ERROR(value,STRAYENDSELECT);
        }
      }
      break;
    }
    /*}}}*/
    case T_ENVIRON: /*{{{*/
    {
      struct Pc epc=pc;

      ++pc.token;
      if (eval(value,"environment variable")->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR) return value;
      if (pass==INTERPRET && value->u.string.character)
      {
        if (putenv(value->u.string.character)==-1)
        {
          Value_destroy(value);
          pc=epc;
          return Value_new_ERROR(value,ENVIRONFAILED,strerror(errno));
        }
      }
      Value_destroy(value);
      break;
    }
    /*}}}*/
    case T_EOL: /*{{{*/
    {
      break;
    }
    /*}}}*/
    funcreturn:
    case T_EQ:
    case T_FNRETURN:
    case T_FNEND: /*{{{*/
    {
      struct Pc *curfn=(struct Pc*)0;
      struct Pc eqpc=pc;
      enum TokenType t=pc.token->type;

      if (pass==DECLARE || pass==COMPILE)
      {
        if (t==T_EQ)
        {
          if ((curfn=popLabel(L_FUNC))==(struct Pc*)0) return Value_new_ERROR(value,STRAYENDEQ);
          if ((eqpc.token->u.type=(curfn->token+1)->u.identifier->defaultType)==V_VOID) return Value_new_ERROR(value,STRAYENDEQ);
        }
        else if (t==T_FNEND)
        {
          if ((curfn=popLabel(L_FUNC))==(struct Pc*)0) return Value_new_ERROR(value,STRAYENDFN);
          if ((eqpc.token->u.type=(curfn->token+1)->u.identifier->defaultType)==V_VOID) return Value_new_ERROR(value,STRAYENDFN);
        }
        else
        {
          if ((curfn=findLabel(L_FUNC))==(struct Pc*)0) return Value_new_ERROR(value,STRAYFNRETURN);
          if ((eqpc.token->u.type=(curfn->token+1)->u.identifier->defaultType)==V_VOID) return Value_new_ERROR(value,STRAYFNRETURN);
        }
      }
      ++pc.token;
      if (eval(value,"return")->type==V_ERROR || Value_retype(value,eqpc.token->u.type)->type==V_ERROR)
      {
        pc=eqpc;
        return value;
      }
      if (pass==INTERPRET) return value;
      else
      {
        Value_destroy(value);
        if (t==T_EQ || t==T_FNEND)
        {
          if (pass==DECLARE) Global_endfunction(&globals,(curfn->token+1)->u.identifier,&pc);
          Auto_funcEnd(&stack);
        }
      }
      break;
    }
    /*}}}*/
    case T_ERASE: /*{{{*/
    {
      ++pc.token;
      while (1)
      {
        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"array variable");
        if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,G_ARRAY,0)==0)
        {
          return Value_new_ERROR(value,REDECLARATION);
        }
        if (pass==INTERPRET)
        {
          Var_destroy(&pc.token->u.identifier->sym->u.var);
        }
        ++pc.token;
        if (pc.token->type==T_COMMA) ++pc.token;
        else break;
      }
      break;
    }
    /*}}}*/
    case T_EXITDO: /*{{{*/
    {
      if (pass==INTERPRET) pc=pc.token->u.exitdo;
      else
      {
        if (pass==COMPILE)
        {
          struct Pc *exitdo;

          if ((exitdo=findLabel(L_DO))==(struct Pc*)0 && (exitdo=findLabel(L_DOWHILE))==(struct Pc*)0) return Value_new_ERROR(value,STRAYEXITDO);
          pc.token->u.exitdo=exitdo->token->u.exitdo;
        }
        ++pc.token;
      }
      break;
    }
    /*}}}*/
    case T_EXITFOR: /*{{{*/
    {
      if (pass==INTERPRET) pc=pc.token->u.exitfor;
      else
      {
        if (pass==COMPILE)
        {
          struct Pc *exitfor;

          if ((exitfor=findLabel(L_FOR))==(struct Pc*)0) return Value_new_ERROR(value,STRAYEXITFOR);
          pc.token->u.exitfor=exitfor->token->u.exitfor;
        }
        ++pc.token;
      }
      break;
    }
    /*}}}*/
    case T_FIELD: /*{{{*/
    {
      long int chn,offset,recLength;
      char *buffer;

      ++pc.token;
      if (pc.token->type==T_CHANNEL) ++pc.token;
      if (eval(value,"channel")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
      chn=value->u.integer;
      Value_destroy(value);
      if (pass==INTERPRET && (recLength=FS_recLength(chn))==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
      if (pc.token->type!=T_COMMA) return Value_new_ERROR(value,MISSINGCOMMA);
      ++pc.token;
      offset=0;
      while (1)
      {
        struct Pc curpc;
        struct Value *l;
        long int width;

        curpc=pc;
        if (eval(value,"width")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        width=value->u.integer;
        Value_destroy(value);
        if (pass==INTERPRET && offset+width>recLength)
        {
          pc=curpc;
          return Value_new_ERROR(value,OUTOFRANGE,"field width");
        }
        if (pc.token->type!=T_AS) return Value_new_ERROR(value,MISSINGAS);
        ++pc.token;
        curpc=pc;
        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"variable");
        if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR,0)==0)
        {
          return Value_new_ERROR(value,REDECLARATION);
        }
        if ((l=lvalue(value))->type==V_ERROR) return value;
        if (pass!=DECLARE && l->type!=V_STRING)
        {
          pc=curpc;
          return Value_new_ERROR(value,TYPEMISMATCH4);
        }
        if (pass==INTERPRET) FS_field(chn,&l->u.string,offset,width);
        offset+=width;
        if (pc.token->type==T_COMMA) ++pc.token;
        else break;
      }
      break;
    }
    /*}}}*/
    case T_FOR: /*{{{*/
    {
      struct Pc forpc=pc;
      struct Pc varpc;
      struct Pc topc;
      struct Pc limitpc;
      struct Value limit,step;

      ++pc.token;
      varpc=pc;
      if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"loop variable");
      if (assign(value)->type==V_ERROR) return value;
      if (pass==INTERPRET)
      {
        struct Value null;

        topc=pc;
        ++pc.token;
        if (eval(&limit,(const char*)0)->type==V_ERROR)
        {
          *value=limit;
          return value;
        }
        Value_retype(&limit,value->type);
        assert(limit.type!=V_ERROR);
        if (pc.token->type==T_STEP) /* STEP x */ /*{{{*/
        {
          struct Pc steppc;

          ++pc.token;
          steppc=pc;
          if (eval(&step,"step")->type==V_ERROR)
          {
            Value_destroy(value);
            *value=step;
            pc=steppc;
            return value;
          }
          Value_retype(&step,value->type);
          assert(step.type!=V_ERROR);
        }
        /*}}}*/
        else /* implicit numeric STEP */ /*{{{*/
        {
          if (value->type==V_INTEGER) VALUE_NEW_INTEGER(&step,1);
          else VALUE_NEW_REAL(&step,1.0);
        }
        /*}}}*/
        if (Value_exitFor(value,&limit,&step)) pc=forpc.token->u.exitfor;
        Value_destroy(&limit);
        Value_destroy(&step);
        Value_destroy(value);
      }
      else
      {
        pushLabel(L_FOR,&forpc);
        pushLabel(L_FOR_VAR,&varpc);
        if (pc.token->type!=T_TO)
        {
          Value_destroy(value);
          return Value_new_ERROR(value,MISSINGTO);
        }
        topc=pc;
        ++pc.token;
        pushLabel(L_FOR_LIMIT,&pc);
        limitpc=pc;
        if (eval(&limit,(const char*)0)==(struct Value*)0)
        {
          Value_destroy(value);
          return Value_new_ERROR(value,MISSINGEXPR,"`to'");
        }
        if (limit.type==V_ERROR)
        {
          Value_destroy(value);
          *value=limit;
          return value;
        }
        if (pass!=DECLARE && VALUE_RETYPE(&limit,varpc.token->u.identifier->sym->u.var.type)->type==V_ERROR)
        {
          Value_destroy(value);
          *value=limit;
          pc=limitpc;
          return value;
        }
        Value_destroy(&limit);
        if (pc.token->type==T_STEP) /* STEP x */ /*{{{*/
        {
          struct Pc steppc;

          ++pc.token;
          steppc=pc;
          if (eval(&step,"`step'")->type==V_ERROR || (pass!=DECLARE && Value_retype(&step,value->type)->type==V_ERROR))
          {
            Value_destroy(value);
            *value=step;
            pc=steppc;
            return value;
          }
        }
        /*}}}*/
        else /* implicit numeric STEP */ /*{{{*/
        {
          VALUE_NEW_INTEGER(&step,1);
          if (pass!=DECLARE && VALUE_RETYPE(&step,value->type)->type==V_ERROR)
          {
            Value_destroy(value);
            *value=step;
            Value_errorPrefix(value,"implicit STEP 1 ");
            return value;
          }
        }
        /*}}}*/
        pushLabel(L_FOR_BODY,&pc);
        Value_destroy(&step);
        Value_destroy(value);
      }
      break;
    }
    /*}}}*/
    case T_GET:
    case T_PUT: /*{{{*/
    {
      int put=pc.token->type==T_PUT;
      long int chn,rec;
      struct Pc errpc;
      
      ++pc.token;
      if (pc.token->type==T_CHANNEL) ++pc.token;
      if (eval(value,"channel")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
      chn=value->u.integer;
      Value_destroy(value);
      if (pc.token->type==T_COMMA)
      {
        ++pc.token;
        errpc=pc;
        if (eval(value,"record number")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        rec=value->u.integer;
        Value_destroy(value);
        if (pass==INTERPRET)
        {
          if (rec<1)
          {
            pc=errpc;
            return Value_new_ERROR(value,OUTOFRANGE,"record number");
          }
          if (FS_seek(chn,rec-1)==-1)
          {
            pc=statementpc;
            return Value_new_ERROR(value,IOERROR,FS_errmsg);
          }
        }
      }
      if (pass==INTERPRET && ((put ? FS_put : FS_get)(chn))==-1)
      {
        pc=statementpc;
        return Value_new_ERROR(value,IOERROR,FS_errmsg);
      }
      break;
    }
    /*}}}*/
    case T_GOSUB: /*{{{*/
    {
      if (pass==INTERPRET)
      {
        if (!runnable && compile(value)->type==V_ERROR) return value;
        pc.token+=2;
        Auto_pushGosubRet(&stack,&pc);
        pc=(pc.token-2)->u.gosubpc;
        Program_trace(&program,&pc,0,1);
      }
      if (pass==DECLARE || pass==COMPILE)
      {
        struct Token *gosubpc=pc.token;

        ++pc.token;
        if (pc.token->type!=T_INTEGER) return Value_new_ERROR(value,MISSINGLINENUMBER);
        if (Program_goLine(&program,pc.token->u.integer,&gosubpc->u.gosubpc)==(struct Pc*)0) return Value_new_ERROR(value,NOSUCHLINE);
        if (pass==COMPILE && Program_scopeCheck(&program,&gosubpc->u.gosubpc,findLabel(L_FUNC))) return Value_new_ERROR(value,UNREACHABLE);
        ++pc.token;
      }
      break;
    }
    /*}}}*/
    case T_RESUME:
    case T_GOTO: /*{{{*/
    {
      if (pass==INTERPRET)
      {
        if (!runnable && compile(value)->type==V_ERROR) return value;
        if (pc.token->type==T_RESUME)
        {
          if (!stack.resumeable) return Value_new_ERROR(value,STRAYRESUME);
          stack.resumeable=0;
        }
        pc=pc.token->u.gotopc;
        Program_trace(&program,&pc,0,1);
      }
      else if (pass==DECLARE || pass==COMPILE)
      {
        struct Token *gotopc=pc.token;

        ++pc.token;
        if (pc.token->type!=T_INTEGER) return Value_new_ERROR(value,MISSINGLINENUMBER);
        if (Program_goLine(&program,pc.token->u.integer,&gotopc->u.gotopc)==(struct Pc*)0) return Value_new_ERROR(value,NOSUCHLINE);
        if (pass==COMPILE && Program_scopeCheck(&program,&gotopc->u.gotopc,findLabel(L_FUNC))) return Value_new_ERROR(value,UNREACHABLE);
        ++pc.token;
      }
      break;
    }
    /*}}}*/
    case T_KILL: /*{{{*/
    {
      ++pc.token;
      if (eval(value,"file name")->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR) return value;
      if (pass==INTERPRET && unlink(value->u.string.character)==-1)
      {
        char *msg=strerror(errno);

        Value_destroy(value);
        pc=statementpc;
        return Value_new_ERROR(value,IOERROR,msg);
      }
      else Value_destroy(value);
      break;
    }
    /*}}}*/
    case T_LET: /*{{{*/
    {
      ++pc.token;
      if (assign(value)->type==V_ERROR) return value;
      if (pass!=INTERPRET) Value_destroy(value);
      break;
    }
    /*}}}*/
    case T_LINEINPUT: /*{{{*/
    {
      int channel=0;
      struct Pc lpc;
      struct Value *l;

      ++pc.token;
      if (pc.token->type==T_CHANNEL) /*{{{*/
      {
        ++pc.token;
        if (eval(value,"channel")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        channel=value->u.integer;
        Value_destroy(value);
        if (pc.token->type!=T_COMMA) return Value_new_ERROR(value,MISSINGCOMMA);
        else ++pc.token;
      }
      /*}}}*/
      /* prompt */ /*{{{*/
      if (pc.token->type==T_STRING)
      {
        if (pass==INTERPRET && channel==0) FS_putString(channel,pc.token->u.string);
        ++pc.token;
        if (pc.token->type!=T_SEMICOLON) return Value_new_ERROR(value,MISSINGSEMICOLON);
        ++pc.token;
      }
      if (pass==INTERPRET && channel==0) FS_flush(channel);
      /*}}}*/
      if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"variable");
      if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR,0)==0)
      {
        return Value_new_ERROR(value,REDECLARATION);
      }
      lpc=pc;
      if (((l=lvalue(value))->type)==V_ERROR) return value;
      if (pass==COMPILE && l->type!=V_STRING)
      {
        pc=lpc;
        return Value_new_ERROR(value,TYPEMISMATCH4);
      }
      if (pass==INTERPRET)
      {
        String_size(&l->u.string,0);
        if (FS_appendToString(channel,&l->u.string,'\n',1)==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
        if (l->u.string.length==0) return Value_new_ERROR(value,IOERROR,"end of file");
        String_size(&l->u.string,l->u.string.length-1);
      }
      break;
    }
    /*}}}*/
    case T_LIST:
    case T_LLIST: /*{{{*/
    {
      struct Pc listpc=pc,from,to;
      int f=0,t=0,channel;

      channel=(pc.token->type==T_LLIST?LPCHANNEL:STDCHANNEL);
      ++pc.token;
      if (pc.token->type==T_INTEGER)
      {
        if (Program_goLine(&program,pc.token->u.integer,&from)==(struct Pc*)0) return Value_new_ERROR(value,NOSUCHLINE);
        f=1;
        ++pc.token;
      }
      if (pc.token->type==T_MINUS)
      {
        ++pc.token;
        if (pc.token->type==T_INTEGER)
        {
          if (Program_goLine(&program,pc.token->u.integer,&to)==(struct Pc*)0) return Value_new_ERROR(value,NOSUCHLINE);
          t=1;
          ++pc.token;
        }
      }
      else if (f==1)
      {
        to=from;
        t=1;
      }
      if (pass==INTERPRET)
      {
        if (!DIRECTMODE)
        {
          pc=listpc;
          return Value_new_ERROR(value,NOTINPROGRAMMODE);
        }
        if (Program_list(&program,channel,f?&from:(struct Pc*)0,t?&to:(struct Pc*)0,value)) return value;
      }
      break;
    }
    /*}}}*/
    case T_LOAD: /*{{{*/
    {
      struct Pc loadpc;

      if (pass==INTERPRET && !DIRECTMODE) return Value_new_ERROR(value,NOTINPROGRAMMODE);
      ++pc.token;
      loadpc=pc;
      if (eval(value,"file name")->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR)
      {
        pc=loadpc;
        return value;
      }
      if (pass==INTERPRET)
      {
        int dev;

        new();
        if ((dev=FS_openin(value->u.string.character))==-1)
        {
          pc=loadpc;
          Value_destroy(value);
          return Value_new_ERROR(value,IOERROR,FS_errmsg);
        }
        Value_destroy(value);
        if (Program_merge(&program,dev,value))
        {
          pc=loadpc;
          return value;
        }
        FS_close(dev);
      }
      else Value_destroy(value);
      break;
    }
    /*}}}*/
    case T_LOCAL: /*{{{*/
    {
      struct Pc *curfn=(struct Pc*)0;

      if (pass==DECLARE || pass==COMPILE)
      {
        if ((curfn=findLabel(L_FUNC))==(struct Pc*)0) return Value_new_ERROR(value,STRAYLOCAL);
      }
      ++pc.token;
      while (1)
      {
        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"variable");
        if (pass==DECLARE || pass==COMPILE)
        {
          struct Symbol *fnsym;

          if (Auto_variable(&stack,pc.token->u.identifier)==0) return Value_new_ERROR(value,ALREADYLOCAL);
          if (pass==DECLARE)
          {
            assert(curfn->token->type==T_DEFFN || curfn->token->type==T_DEFPROC || curfn->token->type==T_SUB);
            fnsym=(curfn->token+1)->u.identifier->sym;
            assert(fnsym);
            fnsym->u.sub.u.def.localTypes=realloc(fnsym->u.sub.u.def.localTypes,sizeof(enum ValueType)*(fnsym->u.sub.u.def.localLength+1));
            fnsym->u.sub.u.def.localTypes[fnsym->u.sub.u.def.localLength]=pc.token->u.identifier->defaultType;
            ++fnsym->u.sub.u.def.localLength;
          }
        }
        ++pc.token;
        if (pc.token->type==T_COMMA) ++pc.token;
        else break;
      }
      break;
    }
    /*}}}*/
    case T_LOCATE: /*{{{*/
    {
      long int line,column;
      struct Pc argpc;

      ++pc.token;
      argpc=pc;
      if (eval(value,"line")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
      line=value->u.integer;
      Value_destroy(value);
      if (pass==INTERPRET && line<1) 
      {
        pc=argpc;
        return Value_new_ERROR(value,OUTOFRANGE,"line");
      }
      if (pc.token->type==T_COMMA) ++pc.token;
      else return Value_new_ERROR(value,MISSINGCOMMA);
      argpc=pc;
      if (eval(value,"column")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
      column=value->u.integer;
      Value_destroy(value);
      if (pass==INTERPRET && column<1) 
      {
        pc=argpc;
        return Value_new_ERROR(value,OUTOFRANGE,"column");
      }
      if (pass==INTERPRET && FS_locate(STDCHANNEL,line,column)==-1)
      {
        pc=statementpc;
        return Value_new_ERROR(value,IOERROR,FS_errmsg);
      }
      break;
    }
    /*}}}*/
    case T_LOOP: /*{{{*/
    {
      struct Pc looppc=pc;
      struct Pc *dopc;

      ++pc.token;
      if (pass==INTERPRET)
      {
        pc=looppc.token->u.dopc;
      }
      if (pass==DECLARE || pass==COMPILE)
      {
        if ((dopc=popLabel(L_DO))==(struct Pc*)0 && (dopc=popLabel(L_DOWHILE))==(struct Pc*)0) return Value_new_ERROR(value,STRAYLOOP);
        looppc.token->u.dopc=*dopc;
        dopc->token->u.exitdo=pc;
      }
      break;
    }
    /*}}}*/
    case T_LOOPUNTIL: /*{{{*/
    {
      struct Pc loopuntilpc=pc;
      struct Pc *dopc;

      ++pc.token;
      if (eval(value,"condition")->type==V_ERROR) return value;
      if (pass==INTERPRET)
      {
        if (Value_isNull(value)) pc=loopuntilpc.token->u.dopc;
        Value_destroy(value);
      }
      if (pass==DECLARE || pass==COMPILE)
      {
        if ((dopc=popLabel(L_DO))==(struct Pc*)0) return Value_new_ERROR(value,STRAYLOOPUNTIL);
        loopuntilpc.token->u.until=*dopc;
        dopc->token->u.exitdo=pc;
      }
      break;
    }
    /*}}}*/
    case T_LSET:
    case T_RSET: /*{{{*/
    {
      struct Value *l;
      struct Pc tmppc;
      int lset=(pc.token->type==T_LSET);

      ++pc.token;
      if (pass==DECLARE)
      {
        if
        (
          ((pc.token+1)->type==T_OP || Auto_find(&stack,pc.token->u.identifier)==0)
          && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR,0)==0
        )
        {
          return Value_new_ERROR(value,REDECLARATION);
        }
      }
      tmppc=pc;
      if ((l=lvalue(value))->type==V_ERROR) return value;
      if (pass==COMPILE && l->type!=V_STRING)
      {
        pc=tmppc;
        return Value_new_ERROR(value,TYPEMISMATCH4);
      }
      if (pc.token->type!=T_EQ) return Value_new_ERROR(value,MISSINGEQ);
      ++pc.token;
      tmppc=pc;
      if (eval(value,"rhs")->type==V_ERROR || (pass!=DECLARE && Value_retype(value,l->type)->type==V_ERROR))
      {
        pc=tmppc;
        return value;
      }
      if (pass==INTERPRET) (lset ? String_lset : String_rset)(&l->u.string,&value->u.string);
      Value_destroy(value);
      break;
    }
    /*}}}*/
    case T_IDENTIFIER: /*{{{*/
    {
      struct Pc here=pc;
      const char *name=pc.token->u.identifier->name,*mid="mid$";

      while (*name && tolower(*name)==*mid) { ++name; ++mid; };
      if (tolower(*name)==*mid) /* mid$(a$,n,m)=b$ */ /*{{{*/
      {
        struct Value *l;
        long int n,m;

        ++pc.token;
        if (pc.token->type!=T_OP) return Value_new_ERROR(value,MISSINGOP);
        ++pc.token;
        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"string variable");
        if (pass==DECLARE)
        {
          if
          (
            ((pc.token+1)->type==T_OP || Auto_find(&stack,pc.token->u.identifier)==0)
            && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR,0)==0
          )
          {
            return Value_new_ERROR(value,REDECLARATION);
          }
        }
        if ((l=lvalue(value))->type==V_ERROR) return value;
        if (pass==COMPILE && l->type!=V_STRING) return Value_new_ERROR(value,TYPEMISMATCH4);
        if (pc.token->type!=T_COMMA) return Value_new_ERROR(value,MISSINGCOMMA);
        ++pc.token;
        if (eval(value,"position")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        n=value->u.integer;
        Value_destroy(value);
        if (pass==INTERPRET && n<1) return Value_new_ERROR(value,OUTOFRANGE,"position");
        if (pc.token->type==T_COMMA)
        {
          ++pc.token;
          if (eval(value,"length")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
          m=value->u.integer;
          if (pass==INTERPRET && m<0) return Value_new_ERROR(value,OUTOFRANGE,"length");
          Value_destroy(value);
        }
        else m=-1;
        if (pc.token->type!=T_CP) return Value_new_ERROR(value,MISSINGCP);
        ++pc.token;
        if (pc.token->type!=T_EQ) return Value_new_ERROR(value,MISSINGEQ);
        ++pc.token;
        if (eval(value,"rhs")->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR) return value;
        if (pass==INTERPRET)
        {
          if (m==-1) m=value->u.string.length;
          String_set(&l->u.string,n-1,&value->u.string,m);
        }
        Value_destroy(value);
      }
      /*}}}*/
      else /* a=b or procedure call */ /*{{{*/
      {
        if (pass==DECLARE)
        {
          if (func(value)->type==V_ERROR) return value;
          else Value_destroy(value);
          if (pc.token->type==T_EQ)
          {
            pc=here;
            if (assign(value)->type==V_ERROR) return value;
            Value_destroy(value);
          }
        }
        else
        {
          if (pass==COMPILE)
          {
            if
            (
              ((pc.token+1)->type==T_OP || Auto_find(&stack,pc.token->u.identifier)==0)
              && Global_find(&globals,pc.token->u.identifier,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR)==0
            ) return Value_new_ERROR(value,UNDECLARED);
          }
          if (pc.token->u.identifier->sym->type==USERFUNCTION)
          {
            func(value);
            if (Value_retype(value,V_VOID)->type==V_ERROR) return value;
            Value_destroy(value);
          }
          else
          {
            if (assign(value)->type==V_ERROR) return value;
            if (pass!=INTERPRET) Value_destroy(value);
          }
        }
      }
      /*}}}*/
      break;
    }
    /*}}}*/
    case T_IF:
    case T_ELSEIFIF: /*{{{*/
    {
      struct Pc ifpc=pc;

      ++pc.token;
      if (eval(value,"condition")->type==V_ERROR) return value;
      if (pc.token->type!=T_THEN)
      {
        Value_destroy(value);
        return Value_new_ERROR(value,MISSINGTHEN);
      }
      ++pc.token;
      if (pass==INTERPRET)
      {
        if (Value_isNull(value)) pc=ifpc.token->u.elsepc;
        Value_destroy(value);
      }
      else
      {
        Value_destroy(value);
        if (pc.token->type==T_EOL)
        {
          pushLabel(L_IF,&ifpc);
        }
        else /* compile single line IF THEN ELSE recursively */ /*{{{*/
        {
          if (statements(value)->type==V_ERROR) return value;
          Value_destroy(value);
          if (pc.token->type==T_ELSE)
          {
            struct Pc elsepc=pc;

            ++pc.token;
            ifpc.token->u.elsepc=pc;
            if (ifpc.token->type==T_ELSEIFIF) (ifpc.token-1)->u.elsepc=pc;
            if (statements(value)->type==V_ERROR) return value;
            Value_destroy(value);
            elsepc.token->u.endifpc=pc;
          }
          else
          {
            ifpc.token->u.elsepc=pc;
            if (ifpc.token->type==T_ELSEIFIF) (ifpc.token-1)->u.elsepc=pc;
          }
        }
        /*}}}*/
      }
      break;
    }
    /*}}}*/
    case T_INPUT: /*{{{*/
    {
      int channel=STDCHANNEL;
      int nl=1;
      int extraprompt=1;
      struct Token *inputdata,*t;
      struct Pc lvaluepc;

      ++pc.token;
      if (pc.token->type==T_CHANNEL) /*{{{*/
      {
        ++pc.token;
        if (eval(value,"channel")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        channel=value->u.integer;
        Value_destroy(value);
        if (pc.token->type!=T_COMMA) return Value_new_ERROR(value,MISSINGCOMMA);
        else ++pc.token;
      }
      /*}}}*/
      if (pc.token->type==T_SEMICOLON) /*{{{*/
      {
        nl=0;
        ++pc.token;
      }
      /*}}}*/
      /* prompt */ /*{{{*/
      if (pc.token->type==T_STRING)
      {
        if (pass==INTERPRET && channel==STDCHANNEL) FS_putString(STDCHANNEL,pc.token->u.string);
        ++pc.token;
        if (pc.token->type==T_COMMA) { ++pc.token; extraprompt=0; }
        else if (pc.token->type==T_SEMICOLON) ++pc.token;
        else extraprompt=0;
      }
      if (pass==INTERPRET && channel==STDCHANNEL && extraprompt)
      {
        FS_putChars(STDCHANNEL,"? ");
      }
      /*}}}*/
      retry:
      if (pass==INTERPRET) /* read input line and tokenise it */ /*{{{*/
      {
        struct String s;

        if (channel==STDCHANNEL) FS_flush(STDCHANNEL);
        String_new(&s);
        if (FS_appendToString(channel,&s,'\n',nl)==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
        if (s.length==0) return Value_new_ERROR(value,IOERROR,"end of file");
        inputdata=t=Token_newData(s.character);
        String_destroy(&s);
      }
      /*}}}*/
      while (1)
      {
        struct Value *l;

        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"variable");
        if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR,0)==0)
        {
          return Value_new_ERROR(value,REDECLARATION);
        }
        lvaluepc=pc;
        if (((l=lvalue(value))->type)==V_ERROR) return value;
        if (pass==INTERPRET)
        {
          if (t->type==T_COMMA || t->type==T_EOL)
          {
            enum ValueType ltype=l->type;

            Value_destroy(l);
            Value_new_null(l,ltype);
          }
          else if (convert(value,l,t))
          {
            pc=lvaluepc;
            if (channel==STDCHANNEL)
            {
              struct String s;

              String_new(&s);
              Value_toString(value,&s,' ',-1,0,0,-1,0);
              FS_putString(STDCHANNEL,&s);
              String_destroy(&s);
              Value_destroy(value);
              FS_putChars(STDCHANNEL,"?? ");
              Token_destroy(inputdata);
              goto retry;
            }
            else
            {
              Token_destroy(inputdata);
              return value;
            }
          }
          else ++t;
          if (pc.token->type==T_COMMA)
          {
            if (t->type==T_COMMA) ++t;
            else
            {
              Token_destroy(inputdata);
              if (channel==STDCHANNEL)
              {
                FS_putChars(STDCHANNEL,"?? ");
                ++pc.token;
                goto retry;
              }
              else
              {
                pc=lvaluepc;
                return Value_new_ERROR(value,MISSINGINPUTDATA);
              }
            }
          }
        }
        if (pc.token->type==T_COMMA) ++pc.token;
        else break;
      }
      if (pass==INTERPRET) Token_destroy(inputdata);
      break;
    }
    /*}}}*/
    case T_MAT: /*{{{*/
    {
      struct Var *var1,*var2,*var3=(struct Var*)0;
      struct Pc oppc;
      enum TokenType op;

      ++pc.token;
      if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"matrix variable");
      if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,G_ARRAY,0)==0)
      {
        return Value_new_ERROR(value,REDECLARATION);
      }
      var1=&pc.token->u.identifier->sym->u.var;
      ++pc.token;
      if (pc.token->type!=T_EQ) return Value_new_ERROR(value,MISSINGEQ);
      ++pc.token;
      if (pc.token->type==T_IDENTIFIER) /* a = b [ +|-|* c ] */ /*{{{*/
      {
        if (pass==COMPILE)
        {
          if
          (
            ((pc.token+1)->type==T_OP || Auto_find(&stack,pc.token->u.identifier)==0)
            && Global_find(&globals,pc.token->u.identifier,G_ARRAY)==0
          ) return Value_new_ERROR(value,UNDECLARED);
        }
        var2=&pc.token->u.identifier->sym->u.var;
        if (pass==INTERPRET && var2->dim!=2) return Value_new_ERROR(value,NOMATRIX,var2->dim);
        if (pass==COMPILE && Value_commonType[var1->type][var2->type]==V_ERROR) return Value_new_typeError(value,var2->type,var1->type);
        ++pc.token;
        if (pc.token->type==T_PLUS || pc.token->type==T_MINUS || pc.token->type==T_MULT)
        {
          oppc=pc;
          op=pc.token->type;
          ++pc.token;
          if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"array variable");
          if (pass==COMPILE)
          {
            if
            (
              ((pc.token+1)->type==T_OP || Auto_find(&stack,pc.token->u.identifier)==0)
              && Global_find(&globals,pc.token->u.identifier,G_ARRAY)==0
            ) return Value_new_ERROR(value,UNDECLARED);
          }
          var3=&pc.token->u.identifier->sym->u.var;
          if (pass==INTERPRET && var3->dim!=2) return Value_new_ERROR(value,NOMATRIX,var3->dim);
          ++pc.token;
        }
        if (pass!=DECLARE)
        {
          if (var3==(struct Var*)0)
          {
            if (Var_mat_assign(var1,var2,value,pass==INTERPRET))
            {
              pc=oppc;
              return value;
            }
          }
          else if (op==T_MULT)
          {
            if (Var_mat_mult(var1,var2,var3,value,pass==INTERPRET))
            {
              pc=oppc;
              return value;
            }
          }
          else if (Var_mat_addsub(var1,var2,var3,op==T_PLUS,value,pass==INTERPRET))
          {
            pc=oppc;
            return value;
          }
        }
      }
      /*}}}*/
      else if (pc.token->type==T_OP) /*{{{*/
      {
        if (var1->type==T_STRING) return Value_new_ERROR(value,TYPEMISMATCH5);
        ++pc.token;
        if (eval(value,"factor")->type==V_ERROR) return value;
        if (pass==COMPILE && Value_commonType[var1->type][value->type]==V_ERROR) return Value_new_typeError(value,var1->type,value->type);
        if (pc.token->type!=T_CP)
        {
          Value_destroy(value);
          return Value_new_ERROR(value,MISSINGCP);
        }
        ++pc.token;
        if (pc.token->type!=T_MULT)
        {
          Value_destroy(value);
          return Value_new_ERROR(value,MISSINGMULT);
        }
        oppc=pc;
        ++pc.token;
        if (pass==COMPILE)
        {
          if
          (
            ((pc.token+1)->type==T_OP || Auto_find(&stack,pc.token->u.identifier)==0)
            && Global_find(&globals,pc.token->u.identifier,G_ARRAY)==0
          )
          {
            Value_destroy(value);
            return Value_new_ERROR(value,UNDECLARED);
          }
        }
        var2=&pc.token->u.identifier->sym->u.var;
        if (pass==INTERPRET && var2->dim!=2)
        {
          Value_destroy(value);
          return Value_new_ERROR(value,NOMATRIX,var2->dim);
        }
        if (pass!=DECLARE && Var_mat_scalarMult(var1,value,var2,pass==INTERPRET))
        {
          pc=oppc;
          return value;
        }
        Value_destroy(value);
        ++pc.token;
      }
      /*}}}*/
      else if (pc.token->type==T_CON || pc.token->type==T_ZER || pc.token->type==T_IDN) /*{{{*/
      {
        op=pc.token->type;
        if (pass==COMPILE && Value_commonType[var1->type][V_INTEGER]==V_ERROR) return Value_new_typeError(value,V_INTEGER,var1->type);
        ++pc.token;
        if (pc.token->type==T_OP)
        {
          unsigned int geometry[2];
          enum ValueType vartype=var1->type;

          ++pc.token;
          if (evalGeometry(value,geometry)) return value;
          if (pc.token->type==T_CP) ++pc.token;
          else return Value_new_ERROR(value,MISSINGCP);
          if (pass==INTERPRET)
          {
            Var_destroy(var1);
            Var_new(var1,vartype,2,geometry);
          }
        }
        if (pass==INTERPRET)
        {
          unsigned int i,j;

          if (var1->dim!=2) return Value_new_ERROR(value,NOMATRIX,var1->dim);
          for (i=1; i<var1->geometry[0]; ++i) for (j=1; j<var1->geometry[1]; ++j)
          {
            int c;

            Value_destroy(&(var1->value[i*var1->geometry[1]+j]));
            switch (op)
            {
              case T_CON: c=1; break;
              case T_ZER: c=0; break;
              case T_IDN: c=(i==j?1:0); break;
              default: assert(0);
            }
            if (var1->type==V_INTEGER) Value_new_INTEGER(&(var1->value[i*var1->geometry[1]+j]),c);
            else Value_new_REAL(&(var1->value[i*var1->geometry[1]+j]),(double)c);
          }
        }
      }
      /*}}}*/
      else if (pc.token->type==T_TRN || pc.token->type==T_INV) /*{{{*/
      {
        op=pc.token->type;
        ++pc.token;
        if (pc.token->type!=T_OP) return Value_new_ERROR(value,MISSINGOP);
        ++pc.token;
        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"matrix variable");
        if (pass==COMPILE)
        {
          if
          (
            ((pc.token+1)->type==T_OP || Auto_find(&stack,pc.token->u.identifier)==0)
            && Global_find(&globals,pc.token->u.identifier,G_ARRAY)==0
          ) return Value_new_ERROR(value,UNDECLARED);
        }
        var2=&pc.token->u.identifier->sym->u.var;
        if (pass==COMPILE && Value_commonType[var1->type][var2->type]==V_ERROR) return Value_new_typeError(value,var2->type,var1->type);
        if (pass==INTERPRET)
        {
          if (var2->dim!=2) return Value_new_ERROR(value,NOMATRIX,var2->dim);
          switch (op)
          {
            case T_TRN: Var_mat_transpose(var1,var2); break;
            case T_INV: if (Var_mat_invert(var1,var2,value)) return value; break;
            default: assert(0);
          }
        }
        ++pc.token;
        if (pc.token->type!=T_CP) return Value_new_ERROR(value,MISSINGCP);
        ++pc.token;
      }
      /*}}}*/
      else return Value_new_ERROR(value,MISSINGEXPR,"matrix");
      break;
    }
    /*}}}*/
    case T_MATINPUT: /*{{{*/
    {
      int channel=STDCHANNEL;

      ++pc.token;
      if (pc.token->type==T_CHANNEL) /*{{{*/
      {
        ++pc.token;
        if (eval(value,"channel")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        channel=value->u.integer;
        Value_destroy(value);
        if (pc.token->type!=T_COMMA) return Value_new_ERROR(value,MISSINGCOMMA);
        else ++pc.token;
      }
      /*}}}*/
      while (1)
      {
        struct Pc lvaluepc;
        struct Var *var;

        lvaluepc=pc;
        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"matrix variable");
        if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,G_ARRAY,0)==0)
        {
          return Value_new_ERROR(value,REDECLARATION);
        }
        var=&pc.token->u.identifier->sym->u.var;
        ++pc.token;
        if (pc.token->type==T_OP)
        {
          unsigned int geometry[2];
          enum ValueType vartype=var->type;

          ++pc.token;
          if (evalGeometry(value,geometry)) return value;
          if (pc.token->type==T_CP) ++pc.token;
          else return Value_new_ERROR(value,MISSINGCP);
          if (pass==INTERPRET)
          {
            Var_destroy(var);
            Var_new(var,vartype,2,geometry);
          }
        }
        if (pass==INTERPRET)
        {
          unsigned int i,j;

          if (var->dim!=2) return Value_new_ERROR(value,NOMATRIX,var->dim);
          for (i=1; i<var->geometry[0]; ++i)
          {
            struct String s;
            struct Token *t,*inputdata;

            if (channel==STDCHANNEL)
            {
              FS_putChars(STDCHANNEL,"? ");
              FS_flush(STDCHANNEL);
            }
            String_new(&s);
            if (FS_appendToString(channel,&s,'\n',1)==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
            if (s.length==0) return Value_new_ERROR(value,IOERROR,"end of file");
            inputdata=t=Token_newData(s.character);
            String_destroy(&s);
            for (j=1; j<var->geometry[1]; ++j)
            {
              if (t->type==T_COMMA || t->type==T_EOL)
              {
                Value_destroy(&(var->value[i*var->geometry[1]+j]));
                Value_new_null(&(var->value[i*var->geometry[1]+j]),var->type);
              }
              else if (convert(value,&(var->value[i*var->geometry[1]+j]),t))
              {
                Token_destroy(inputdata);
                pc=lvaluepc;
                return value;
              }
              else ++t;
              if (t->type==T_COMMA) ++t;
            }
            Token_destroy(inputdata);
          }
        }
        if (pc.token->type==T_COMMA) ++pc.token;
        else break;
      }
      break;
    }
    /*}}}*/
    case T_MATPRINT: /*{{{*/
    {
      int chn=STDCHANNEL;
      int printusing=0;
      struct Value usingval;
      struct String *using;
      size_t usingpos=0;
      int notfirst=0;

      ++pc.token;
      if (chn==STDCHANNEL && pc.token->type==T_CHANNEL) /*{{{*/
      {
        ++pc.token;
        if (eval(value,"channel")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        chn=value->u.integer;
        Value_destroy(value);
        if (pc.token->type==T_COMMA) ++pc.token;
      }
      /*}}}*/
      if (pc.token->type==T_USING) /*{{{*/
      {
        printusing=1;
        ++pc.token;
        if (eval(&usingval,(const char*)0)==(struct Value*)0) return Value_new_ERROR(value,MISSINGEXPR,"format string");
        else if (usingval.type==V_ERROR || Value_retype(&usingval,V_STRING)->type==V_ERROR)
        {
          *value=usingval;
          return value;
        }
        if (pc.token->type==T_SEMICOLON) ++pc.token;
        else
        {
          Value_destroy(&usingval);
          return Value_new_ERROR(value,MISSINGSEMICOLON);
        }
      }
      /*}}}*/
      else Value_new_STRING(&usingval);
      using=&usingval.u.string;
      while (1)
      {
        struct Var *var;
        int zoned=1;

        if (pc.token->type!=T_IDENTIFIER)
        {
          Value_destroy(&usingval);
          return Value_new_ERROR(value,MISSINGIDENTIFIER,"array variable");
        }
        if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,G_ARRAY,0)==0)
        {
          Value_destroy(&usingval);
          return Value_new_ERROR(value,REDECLARATION);
        }
        var=&pc.token->u.identifier->sym->u.var;
        ++pc.token;
        if (pc.token->type==T_SEMICOLON) zoned=0;
        if (pass==INTERPRET)
        {
          unsigned int i,j;

          if (var->dim!=2) return Value_new_ERROR(value,NOMATRIX,var->dim);
          if (notfirst && FS_putChar(chn,'\n')==-1)
          {
            Value_destroy(&usingval);
            return Value_new_ERROR(value,IOERROR,FS_errmsg);
          }
          for (i=1; i<var->geometry[0]; ++i)
          {
            for (j=1; j<var->geometry[1]; ++j)
            {
              struct String s;

              String_new(&s);
              Value_clone(value,&(var->value[i*var->geometry[1]+j]));
              if (Value_toStringUsing(value,&s,&usingval.u.string,&usingpos)->type==V_ERROR)
              {
                Value_destroy(&usingval);
                Value_destroy(value);
                String_destroy(&s);
                return value;
              }
              Value_destroy(value);
              if (FS_putString(chn,&s)==-1)
              {
                Value_destroy(&usingval);
                String_destroy(&s);
                return Value_new_ERROR(value,IOERROR,FS_errmsg);
              }
              String_destroy(&s);
              if (!printusing && zoned) FS_nextcol(chn);
            }
            if (FS_putChar(chn,'\n')==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
          }
        }
        if (pc.token->type==T_COMMA || pc.token->type==T_SEMICOLON) ++pc.token;
        else break;
        notfirst=1;
      }
      Value_destroy(&usingval);
      if (pass==INTERPRET)
      {
        if (FS_flush(chn)==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
      }
      break;
    }
    /*}}}*/
    case T_MATREAD: /*{{{*/
    {
      ++pc.token;
      while (1)
      {
        struct Pc lvaluepc;
        struct Var *var;

        lvaluepc=pc;
        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"matrix variable");
        if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,G_ARRAY,0)==0)
        {
          return Value_new_ERROR(value,REDECLARATION);
        }
        var=&pc.token->u.identifier->sym->u.var;
        ++pc.token;
        if (pc.token->type==T_OP)
        {
          unsigned int geometry[2];
          enum ValueType vartype=var->type;

          ++pc.token;
          if (evalGeometry(value,geometry)) return value;
          if (pc.token->type==T_CP) ++pc.token;
          else return Value_new_ERROR(value,MISSINGCP);
          if (pass==INTERPRET)
          {
            Var_destroy(var);
            Var_new(var,vartype,2,geometry);
          }
        }
        if (pass==INTERPRET)
        {
          unsigned int i,j;

          if (var->dim!=2) return Value_new_ERROR(value,NOMATRIX,var->dim);
          for (i=1; i<var->geometry[0]; ++i) for (j=1; j<var->geometry[1]; ++j)
          {
            if (dataread(value,&(var->value[i*var->geometry[1]+j])))
            {
              pc=lvaluepc;
              return value;
            }
          }
        }
        if (pc.token->type==T_COMMA) ++pc.token;
        else break;
      }
      break;
    }
    /*}}}*/
    case T_NAME: /*{{{*/
    {
      struct Pc namepc=pc;
      struct Value old;
      int res,reserrno;

      ++pc.token;
      if (eval(value,"old file name")->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR) return value;
      if (pc.token->type!=T_AS)
      {
        Value_destroy(value);
        return Value_new_ERROR(value,MISSINGAS);
      }
      old=*value;
      ++pc.token;
      if (eval(value,"new file name")->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR)
      {
        Value_destroy(&old);
        return value;
      }
      if (pass==INTERPRET)
      {
        res=rename(old.u.string.character,value->u.string.character);
        reserrno=errno;
      }
      Value_destroy(&old);
      Value_destroy(value);
      if (pass==INTERPRET && res==-1)
      {
        pc=namepc;
        return Value_new_ERROR(value,IOERROR,strerror(reserrno));
      }
      break;
    }
    /*}}}*/
    case T_NEW: /*{{{*/
    {
      if (pass==INTERPRET)
      {
        if (!DIRECTMODE) return Value_new_ERROR(value,NOTINPROGRAMMODE);
        new();
      }
      ++pc.token;
      break;
    }
    /*}}}*/
    case T_NEXT: /*{{{*/
    {
      struct Next **next=&pc.token->u.next;
      int level=0;

      if (pass==INTERPRET) /*{{{*/
      {
        struct Value *l,inc;
        struct Pc savepc;

        ++pc.token;
        while (1)
        {
          /* get variable lvalue */ /*{{{*/
          savepc=pc;
          pc=(*next)[level].var;
          if ((l=lvalue(value))->type==V_ERROR) return value;
          pc=savepc;
          /*}}}*/
          /* get limit value and increment */ /*{{{*/
          savepc=pc;
          pc=(*next)[level].limit;
          if (eval(value,"limit")->type==V_ERROR) return value;
          Value_retype(value,l->type);
          assert(value->type!=V_ERROR);
          if (pc.token->type==T_STEP)
          {
            ++pc.token;
            if (eval(&inc,"step")->type==V_ERROR)
            {
              Value_destroy(value);
              *value=inc;
              return value;
            }
          }
          else VALUE_NEW_INTEGER(&inc,1);
          VALUE_RETYPE(&inc,l->type);
          assert(inc.type!=V_ERROR);
          pc=savepc;
          /*}}}*/
          Value_add(l,&inc,1);
          if (Value_exitFor(l,value,&inc))
          {
            Value_destroy(value);
            Value_destroy(&inc);
            if (pc.token->type==T_IDENTIFIER)
            {
              if (lvalue(value)->type==V_ERROR) return value;
              if (pc.token->type==T_COMMA) { ++pc.token; ++level; }
              else break;
            }
            else break;
          }
          else
          {
            pc=(*next)[level].body;
            Value_destroy(value);
            Value_destroy(&inc);
            break;
          }
        }
      }
      /*}}}*/
      else /*{{{*/
      {
        struct Pc *inc,*body;

        ++pc.token;
        while (1)
        {
          if ((body=popLabel(L_FOR_BODY))==(struct Pc*)0) return Value_new_ERROR(value,STRAYNEXT);
          if (level) *next=realloc(*next,sizeof(struct Next)*(level+1));
          (*next)[level].body=*body;
          (*next)[level].limit=*popLabel(L_FOR_LIMIT);
          (*next)[level].var=*popLabel(L_FOR_VAR);
          (*next)[level].fr=*popLabel(L_FOR);
          if (pc.token->type==T_IDENTIFIER)
          {
            if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR,0)==0)
            {
              return Value_new_ERROR(value,REDECLARATION);
            }
            if (lvalue(value)->type==V_ERROR) return value;
            if (pc.token->type==T_COMMA) { ++pc.token; ++level; }
            else break;
          }
          else break;
        }
        while (level>=0) (*next)[level--].fr.token->u.exitfor=pc;
      }
      /*}}}*/
      break;
    }
    /*}}}*/
    case T_ON: /*{{{*/
    {
      struct On *on=&pc.token->u.on;

      ++pc.token;
      if (eval(value,"selection")->type==V_ERROR) return value;
      if (Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
      if (pass==INTERPRET)
      {
        struct Pc newpc;

        if (value->u.integer>0 && value->u.integer<on->pcLength)
        {
          newpc=on->pc[value->u.integer];
        }
        else newpc=on->pc[0];
        if (pc.token->type==T_GOTO) pc=newpc;
        else
        {
          pc=on->pc[0];
          Auto_pushGosubRet(&stack,&pc);
          pc=newpc;
        }
        Program_trace(&program,&pc,0,1);
      }
      else if (pass==DECLARE || pass==COMPILE)
      {
        Value_destroy(value);
        if (pc.token->type!=T_GOTO && pc.token->type!=T_GOSUB) return Value_new_ERROR(value,MISSINGGOTOSUB);
        ++pc.token;
        on->pcLength=1;
        while (1)
        {
          on->pc=realloc(on->pc,sizeof(struct Pc)*++on->pcLength);
          if (pc.token->type!=T_INTEGER) return Value_new_ERROR(value,MISSINGLINENUMBER);
          if (Program_goLine(&program,pc.token->u.integer,&on->pc[on->pcLength-1])==(struct Pc*)0) return Value_new_ERROR(value,NOSUCHLINE);
          if (pass==COMPILE && Program_scopeCheck(&program,&on->pc[on->pcLength-1],findLabel(L_FUNC))) return Value_new_ERROR(value,UNREACHABLE);
          ++pc.token;
          if (pc.token->type==T_COMMA) ++pc.token;
          else break;
        }
        on->pc[0]=pc;
      }
      break;
    }
    /*}}}*/
    case T_ONERROR: /*{{{*/
    {
      if (DIRECTMODE) return Value_new_ERROR(value,NOTINDIRECTMODE);
      ++pc.token;
      if (pass==INTERPRET)
      {
        stack.onerror=pc;
        Program_nextLine(&program,&pc);
      }
      else goto more;
      break;
    }
    /*}}}*/
    case T_ONERRORGOTO0: /*{{{*/
    {
      if (DIRECTMODE) return Value_new_ERROR(value,NOTINDIRECTMODE);
      if (pass==INTERPRET)
      {
        stack.onerror.line=-1;
        if (stack.resumeable)
        {
          pc=stack.erpc;
          return Value_clone(value,&stack.err);
        }
      }
      ++pc.token;
      break;
    }
    /*}}}*/
    case T_ONERROROFF: /*{{{*/
    {
      if (DIRECTMODE) return Value_new_ERROR(value,NOTINDIRECTMODE);
      if (pass==INTERPRET) stack.onerror.line=-1;
      ++pc.token;
      break;
    }
    /*}}}*/
    case T_OPEN: /*{{{*/
    {
      int inout=-1;
      long int channel;
      long int recLength;
      struct Pc errpc;
      struct Value recLengthValue;

      ++pc.token;
      errpc=pc;
      if (eval(value,"mode or file")->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR) return value;
      if (pc.token->type==T_FOR) /* parse ANSI syntax */ /*{{{*/
      {
        struct Value channelValue;

        ++pc.token;
        if (pc.token->type==T_OUTPUT) inout=1;
        else if (pc.token->type==T_INPUT) inout=0;
        else if (pc.token->type==T_RANDOM) inout=3;
        else
        {
          Value_destroy(value);
          return Value_new_ERROR(value,MISSINGMODE);
        }
        ++pc.token;
        if (pc.token->type!=T_AS)
        {
          Value_destroy(value);
          return Value_new_ERROR(value,MISSINGAS);
        }
        ++pc.token;
        if (pc.token->type!=T_FILE)
        {
          Value_destroy(value);
          return Value_new_ERROR(value,MISSINGFILE);
        }
        ++pc.token;
        if (pc.token->type==T_CHANNEL) ++pc.token;
        errpc=pc;
        if (eval(&channelValue,"channel")->type==V_ERROR || Value_retype(&channelValue,V_INTEGER)->type==V_ERROR)
        {
          pc=errpc;
          Value_destroy(value);
          *value=channelValue;
          return value;
        }
        channel=channelValue.u.integer;
        Value_destroy(&channelValue);
        if (inout==3)
        {
          if (pc.token->type!=T_IDENTIFIER || cistrcmp(pc.token->u.identifier->name,"len"))
          {
            Value_destroy(value);
            return Value_new_ERROR(value,MISSINGLEN);
          }
          ++pc.token;
          if (pc.token->type!=T_EQ)
          {
            Value_destroy(value);
            return Value_new_ERROR(value,MISSINGEQ);
          }
          ++pc.token;
          errpc=pc;
          if (eval(&recLengthValue,"record length")->type==V_ERROR || Value_retype(&recLengthValue,V_INTEGER)->type==V_ERROR)
          {
            Value_destroy(value);
            *value=recLengthValue;
            return value;
          }
          recLength=recLengthValue.u.integer;
          Value_destroy(&recLengthValue);
          if (pass==INTERPRET && recLength<=0)
          {
            Value_destroy(value);
            pc=errpc;
            return Value_new_ERROR(value,OUTOFRANGE,"record length");
          }
        }
      }
      /*}}}*/
      else /* parse MBASIC syntax */ /*{{{*/
      {
        if (value->u.string.length>=1)
        {
          switch (tolower(value->u.string.character[0]))
          {
            case 'i': inout=0; break;
            case 'o': inout=1; break;
            case 'r': inout=3; break;
          }
        }
        Value_destroy(value);
        if (pass==INTERPRET && inout==-1) 
        {
          pc=errpc;
          return Value_new_ERROR(value,BADMODE);
        }
        if (pc.token->type!=T_COMMA) return Value_new_ERROR(value,MISSINGCOMMA);
        ++pc.token;
        if (pc.token->type==T_CHANNEL) ++pc.token;
        errpc=pc;
        if (eval(value,"channel")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR)
        {
          pc=errpc;
          return value;
        }
        channel=value->u.integer;
        Value_destroy(value);
        if (pass==INTERPRET && channel<0) return Value_new_ERROR(value,OUTOFRANGE,"channel");
        if (pc.token->type!=T_COMMA) return Value_new_ERROR(value,MISSINGCOMMA);
        ++pc.token;
        if (eval(value,"file name")->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR) return value;
        if (inout==3)
        {
          if (pc.token->type!=T_COMMA)
          {
            Value_destroy(value);
            return Value_new_ERROR(value,MISSINGCOMMA);
          }
          ++pc.token;
          errpc=pc;
          if (eval(&recLengthValue,"record length")->type==V_ERROR || Value_retype(&recLengthValue,V_INTEGER)->type==V_ERROR)
          {
            Value_destroy(value);
            *value=recLengthValue;
            return value;
          }
          recLength=recLengthValue.u.integer;
          Value_destroy(&recLengthValue);
          if (pass==INTERPRET && recLength<=0)
          {
            Value_destroy(value);
            pc=errpc;
            return Value_new_ERROR(value,OUTOFRANGE,"record length");
          }
        }
      }
      /*}}}*/
      /* open file with name value */ /*{{{*/
      if (pass==INTERPRET)
      {
        int res=-1;

        if (inout==0) res=FS_openinChn(channel,value->u.string.character);
        else if (inout==1) res=FS_openoutChn(channel,value->u.string.character);
        else if (inout==3) res=FS_openrandomChn(channel,value->u.string.character,recLength);
        if (res==-1)
        {
          pc=statementpc;
          Value_destroy(value);
          return Value_new_ERROR(value,IOERROR,FS_errmsg);
        }
      }
      /*}}}*/
      Value_destroy(value);
      break;
    }
    /*}}}*/
    case T_PRINT:
    case T_LPRINT: /*{{{*/
    {
      int nl=1;
      int chn=(pc.token->type==T_PRINT?STDCHANNEL:LPCHANNEL);
      int printusing=0;
      struct Value usingval;
      struct String *using;
      size_t usingpos=0;

      ++pc.token;
      if (chn==STDCHANNEL && pc.token->type==T_CHANNEL) /*{{{*/
      {
        ++pc.token;
        if (eval(value,"channel")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        chn=value->u.integer;
        Value_destroy(value);
        if (pc.token->type==T_COMMA) ++pc.token;
      }
      /*}}}*/
      if (pc.token->type==T_USING) /*{{{*/
      {
        printusing=1;
        ++pc.token;
        if (eval(&usingval,"format string")->type==V_ERROR || Value_retype(&usingval,V_STRING)->type==V_ERROR)
        {
          *value=usingval;
          return value;
        }
        if (pc.token->type!=T_SEMICOLON)
        {
          Value_destroy(&usingval);
          return Value_new_ERROR(value,MISSINGCOMMA);
        }
        else ++pc.token;
      }
      /*}}}*/
      else Value_new_STRING(&usingval);
      using=&usingval.u.string;
      while (1)
      {
        if (eval(value,(const char*)0)) /*{{{*/
        {
          if (value->type==V_ERROR)
          {
            Value_destroy(&usingval);
            return value;
          }
          if (pass==INTERPRET)
          {
            struct String s;

            String_new(&s);
            if (Value_toStringUsing(value,&s,&usingval.u.string,&usingpos)->type==V_ERROR)
            {
              Value_destroy(&usingval);
              String_destroy(&s);
              return value;
            }
            if (FS_putString(chn,&s)==-1)
            {
              Value_destroy(&usingval);
              Value_destroy(value);
              String_destroy(&s);
              return Value_new_ERROR(value,IOERROR,FS_errmsg);
            }
            String_destroy(&s);
          }
          Value_destroy(value);
          nl=1;
        }
        /*}}}*/
        else if (pc.token->type==T_TAB || pc.token->type==T_SPC) /*{{{*/
        {
          int tab=pc.token->type==T_TAB;

          ++pc.token;
          if (pc.token->type!=T_OP)
          {
            Value_destroy(&usingval);
            return Value_new_ERROR(value,MISSINGOP);
          }
          ++pc.token;
          if (eval(value,"format")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR)
          {
            Value_destroy(&usingval);
            return value;
          }
          if (pass==INTERPRET)
          {
            int s=value->u.integer;
            int r=0;

            if (tab) r=FS_tab(chn,s);
            else while (s-->0 && (r=FS_putChar(chn,' '))!=-1);
            if (r==-1)
            {
              Value_destroy(&usingval);
              Value_destroy(value);
              return Value_new_ERROR(value,IOERROR,FS_errmsg);
            }
          }
          Value_destroy(value);
          if (pc.token->type!=T_CP)
          {
            Value_destroy(&usingval);
            return Value_new_ERROR(value,MISSINGCP);
          }
          ++pc.token;
          nl=1;
        }
        /*}}}*/
        else if (pc.token->type==T_SEMICOLON) /*{{{*/
        {
          ++pc.token;
          nl=0;
        }
        /*}}}*/
        else if (pc.token->type==T_COMMA) /*{{{*/
        {
          ++pc.token;
          if (pass==INTERPRET && !printusing) FS_nextcol(chn);
          nl=0;
        }
        /*}}}*/
        else break;
        if (pass==INTERPRET && FS_flush(chn)==-1)
        {
          Value_destroy(&usingval);
          return Value_new_ERROR(value,IOERROR,FS_errmsg);
        }
      }
      Value_destroy(&usingval);
      if (pass==INTERPRET)
      {
        if (nl && FS_putChar(chn,'\n')==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
        if (FS_flush(chn)==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
      }
      break;
    }
    /*}}}*/
    case T_QUOTE: /*{{{*/
    {
      ++pc.token;
      break;
    }
    /*}}}*/
    case T_RANDOMIZE: /*{{{*/
    {
      struct Pc argpc;

      ++pc.token;
      argpc=pc;
      if (eval(value,(const char*)0))
      {
        Value_retype(value,V_INTEGER);
        if (value->type==V_ERROR)
        {
          pc=argpc;
          Value_destroy(value);
          return Value_new_ERROR(value,MISSINGEXPR,"seed");
        }
        if (pass==INTERPRET) srand(pc.token->u.integer);
        Value_destroy(value);
      }
      else srand(getpid()^time((time_t*)0));
      break;
    }
    /*}}}*/
    case T_READ: /*{{{*/
    {
      ++pc.token;
      while (1)
      {
        struct Value *l;
        struct Pc lvaluepc;

        lvaluepc=pc;
        if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"variable");
        if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR,0)==0)
        {
          return Value_new_ERROR(value,REDECLARATION);
        }
        if ((l=lvalue(value))->type==V_ERROR) return value;
        if (pass==INTERPRET && dataread(value,l))
        {
          pc=lvaluepc;
          return value;
        }
        if (pc.token->type==T_COMMA) ++pc.token;
        else break;
      }
      break;
    }
    /*}}}*/
    case T_REM: /*{{{*/
    {
      ++pc.token;
      break;
    }
    /*}}}*/
    case T_RENUM: /*{{{*/
    {
      int first=10,inc=10;

      ++pc.token;
      if (pc.token->type==T_INTEGER)
      {
        first=pc.token->u.integer;
        ++pc.token;
        if (pc.token->type==T_COMMA)
        {
          ++pc.token;
          if (pc.token->type!=T_INTEGER) return Value_new_ERROR(value,MISSINGINCREMENT);
          inc=pc.token->u.integer;
          ++pc.token;
        }
      }
      if (pass==INTERPRET)
      {
        if (!DIRECTMODE) return Value_new_ERROR(value,NOTINPROGRAMMODE);
        Program_renum(&program,first,inc);
        runnable=0;
      }
      break;
    }
    /*}}}*/
    case T_REPEAT: /*{{{*/
    {
      if (pass==DECLARE || pass==COMPILE) pushLabel(L_REPEAT,&pc);
      ++pc.token;
      break;
    }
    /*}}}*/
    case T_RESTORE: /*{{{*/
    {
      struct Token *restorepc=pc.token;

      if (pass==INTERPRET) curdata=pc.token->u.restore;
      ++pc.token;
      if (pc.token->type==T_INTEGER)
      {
        if (pass==COMPILE && Program_dataLine(&program,pc.token->u.integer,&restorepc->u.restore)==(struct Pc*)0) return Value_new_ERROR(value,NOSUCHDATALINE);
        ++pc.token;
      }
      else if (pass==COMPILE) restorepc->u.restore=begindata;
      break;
    }
    /*}}}*/
    case T_RETURN: /*{{{*/
    {
      if (pass==DECLARE || pass==COMPILE) ++pc.token;
      if (pass==INTERPRET)
      {
        if (Auto_gosubReturn(&stack,&pc)) Program_trace(&program,&pc,0,1);
        else return Value_new_ERROR(value,STRAYRETURN);
      }
      break;
    }
    /*}}}*/
    case T_RUN: /*{{{*/
    {
      struct Pc argpc,begin;

      stack.resumeable=0;
      ++pc.token;
      argpc=pc;
      if (pc.token->type==T_INTEGER)
      {
        if (Program_goLine(&program,pc.token->u.integer,&begin)==(struct Pc*)0) return Value_new_ERROR(value,NOSUCHLINE);
        if (pass==COMPILE && Program_scopeCheck(&program,&begin,findLabel(L_FUNC))) return Value_new_ERROR(value,UNREACHABLE);
        ++pc.token;
      }
      else if (eval(value,(const char*)0))
      {
        if (value->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR)
        {
          pc=argpc;
          return value;
        }
        else if (pass==INTERPRET)
        {
          int chn;
          struct Program newprogram;

          if ((chn=FS_openin(value->u.string.character))==-1)
          {
            pc=argpc;
            Value_destroy(value);
            return Value_new_ERROR(value,IOERROR,FS_errmsg);
          }
          Value_destroy(value);
          Program_new(&newprogram);
          if (Program_merge(&newprogram,chn,value))
          {
            pc=argpc;
            Program_destroy(&newprogram);
            return value;
          }
          FS_close(chn);
          new();
          Program_destroy(&program);
          program=newprogram;
          Program_beginning(&program,&begin);
        }
        else Value_destroy(value);
      }
      else Program_beginning(&program,&begin);
      if (pass==INTERPRET)
      {
        if (!runnable && compile(value)->type==V_ERROR) return value;
        pc=begin;
        Global_clear(&globals);
        Program_trace(&program,&pc,0,1);
      }
      break;
    }
    /*}}}*/
    case T_SAVE: /*{{{*/
    {
      struct Pc loadpc;

      if (pass==INTERPRET && !DIRECTMODE) return Value_new_ERROR(value,NOTINPROGRAMMODE);
      ++pc.token;
      loadpc=pc;
      if (eval(value,"file name")->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR)
      {
        pc=loadpc;
        return value;
      }
      if (pass==INTERPRET)
      {
        int chn;

        if ((chn=FS_openout(value->u.string.character,0))==-1)
        {
          pc=loadpc;
          Value_destroy(value);
          return Value_new_ERROR(value,IOERROR,FS_errmsg);
        }
        FS_width(chn,0);
        Value_destroy(value);
        if (Program_list(&program,chn,(struct Pc*)0,(struct Pc*)0,value))
        {
          pc=loadpc;
          return value;
        }
        FS_close(chn);
      }
      else Value_destroy(value);
      break;
    }
    /*}}}*/
    case T_SELECTCASE: /*{{{*/
    {
      if (pass==DECLARE || pass==COMPILE) pushLabel(L_SELECTCASE,&pc);
      ++pc.token;
      if (eval(value,"selector")->type==V_ERROR) return value;
      if (pass==DECLARE || pass==COMPILE)
      {
        statementpc.token->u.selectcase->type=value->type;
        statementpc.token->u.selectcase->nextcasevalue.line=-1;
      }
      else
      {
        struct Pc casevaluepc;
        int match=0;

        pc=casevaluepc=statementpc.token->u.selectcase->nextcasevalue;
        do
        {
          ++pc.token;
          switch (casevaluepc.token->type)
          {
            case T_CASEVALUE: /*{{{*/
            {
              do
              {
                struct Value casevalue1;

                if (eval(&casevalue1,"case")->type==V_ERROR)
                {
                  Value_destroy(value);
                  *value=casevalue1;
                  return value;
                }
                Value_retype(&casevalue1,statementpc.token->u.selectcase->type);
                assert(casevalue1.type!=V_ERROR);
                if (pc.token->type==T_TO) /* match range */ /*{{{*/
                {
                  struct Value casevalue2;

                  ++pc.token;
                  if (eval(&casevalue2,"case")->type==V_ERROR)
                  {
                    Value_destroy(&casevalue1);
                    Value_destroy(value);
                    *value=casevalue2;
                    return value;
                  }
                  Value_retype(&casevalue2,statementpc.token->u.selectcase->type);
                  assert(casevalue2.type!=V_ERROR);
                  switch (casevalue1.type)
                  {
                    case V_INTEGER:
                    {
                      if (!match) match=(value->u.integer>=casevalue1.u.integer && value->u.integer<=casevalue2.u.integer);
                      break;
                    }
                    case V_REAL:
                    {
                      if (!match) match=(value->u.real>=casevalue1.u.real && value->u.real<=casevalue2.u.real);
                      break;
                    }
                    case V_STRING:
                    {
                      if (!match) match=String_cmp(&value->u.string,&casevalue1.u.string)>=0 && String_cmp(&value->u.string,&casevalue2.u.string)<=0;
                      break;
                    }
                    default: assert(0);
                  }
                  Value_destroy(&casevalue2);
                }
                /*}}}*/
                else /* match value */ /*{{{*/
                {
                  switch (casevalue1.type)
                  {
                    case V_INTEGER:
                    {
                      if (!match) match=(value->u.integer==casevalue1.u.integer);
                      break;
                    }
                    case V_REAL:
                    {
                      if (!match) match=(value->u.real==casevalue1.u.real);
                      break;
                    }
                    case V_STRING:
                    {
                      if (!match) match=(String_cmp(&value->u.string,&casevalue1.u.string)==0);
                      break;
                    }
                    default: assert(0);
                  }
                }
                /*}}}*/
                Value_destroy(&casevalue1);
                if (pc.token->type==T_COMMA) ++pc.token;
                else break;
              } while (1);
              break;
            }
            /*}}}*/
            case T_CASEELSE: /*{{{*/
            {
              match=1;
              break;
            }
            /*}}}*/
            default: assert(0);
          }
          if (!match)
          {
            if (casevaluepc.token->u.casevalue->nextcasevalue.line!=-1)
            {
              pc=casevaluepc=casevaluepc.token->u.casevalue->nextcasevalue;
            }
            else
            {
              pc=statementpc.token->u.selectcase->endselect;
              break;
            }
          }
        } while (!match);
      }
      Value_destroy(value);
      break;
    }
    /*}}}*/
    case T_SHELL: /*{{{*/
    {
      pid_t pid;
      int status;

      ++pc.token;
      if (eval(value,(const char*)0))
      {
        if (value->type==V_ERROR || Value_retype(value,V_STRING)->type==V_ERROR) return value;
        if (pass==INTERPRET) switch (pid=fork())
        {
          case -1: return Value_new_ERROR(value,FORKFAILED,strerror(errno));
          case 0:
          {
            execl("/bin/sh","sh","-c",value->u.string.character,(const char*)0);
            exit(127);
          }
          default:
          {
            while (waitpid(pid,&status,0)==-1 && errno!=EINTR);
          }
        }
        Value_destroy(value);
      }
      else
      {
        if (pass==INTERPRET) switch (pid=fork())
        {
          case -1: return Value_new_ERROR(value,FORKFAILED,strerror(errno));
          case 0:
          {
            const char *shell;

            shell=getenv("SHELL");
            if (shell==(const char*)0) shell="/bin/sh";
            execl(shell,(strrchr(shell,'/') ? strrchr(shell,'/')+1 : shell),(const char*)0);
            exit(127);
          }
          default:
          {
            while (waitpid(pid,&status,0)==-1 && errno!=EINTR);
          }
        }
      }
      break;
    }
    /*}}}*/
    case T_STOP: /*{{{*/
    {
      if (pass==INTERPRET) stop=1;
      else ++pc.token;
      break;
    }
    /*}}}*/
    case T_SWAP: /*{{{*/
    {
      struct Value *l1,*l2;
      struct Pc lvaluepc;

      ++pc.token;
      lvaluepc=pc;
      if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"variable");
      if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR,0)==0)
      {
        return Value_new_ERROR(value,REDECLARATION);
      }
      if ((l1=lvalue(value))->type==V_ERROR) return value;
      if (pc.token->type==T_COMMA) ++pc.token;
      else return Value_new_ERROR(value,MISSINGCOMMA);
      lvaluepc=pc;
      if (pc.token->type!=T_IDENTIFIER) return Value_new_ERROR(value,MISSINGIDENTIFIER,"variable");
      if (pass==DECLARE && Global_variable(&globals,pc.token->u.identifier,pc.token->u.identifier->defaultType,(pc.token+1)->type==T_OP?G_ARRAY:G_SCALAR,0)==0)
      {
        return Value_new_ERROR(value,REDECLARATION);
      }
      if ((l2=lvalue(value))->type==V_ERROR) return value;
      if (l1->type!=l2->type)
      {
        pc=lvaluepc;
        return Value_new_typeError(value,l2->type,l1->type);
      }
      if (pass==INTERPRET)
      {
        struct Value foo;

        foo=*l1;
        *l1=*l2;
        *l2=foo;
      }
      break;
    }
    /*}}}*/
    case T_SYSTEM: /*{{{*/
    {
      ++pc.token;
      if (pass==INTERPRET) { bas_exit(); exit(0); }
      break;
    }
    /*}}}*/
    case T_TROFF: /*{{{*/
    {
      ++pc.token;
      program.trace=0;
      break;
    }
    /*}}}*/
    case T_TRON: /*{{{*/
    {
      ++pc.token;
      program.trace=1;
      break;
    }
    /*}}}*/
    case T_UNNUM: /*{{{*/
    {
      ++pc.token;
      if (pass==INTERPRET)
      {
        if (!DIRECTMODE) return Value_new_ERROR(value,NOTINPROGRAMMODE);
        Program_unnum(&program);
        runnable=0;
      }
      break;
    }
    /*}}}*/
    case T_UNTIL: /*{{{*/
    {
      struct Pc untilpc=pc;
      struct Pc *repeatpc;

      ++pc.token;
      if (eval(value,"condition")->type==V_ERROR) return value;
      if (pass==INTERPRET)
      {
        if (Value_isNull(value)) pc=untilpc.token->u.until;
        Value_destroy(value);
      }
      if (pass==DECLARE || pass==COMPILE)
      {
        if ((repeatpc=popLabel(L_REPEAT))==(struct Pc*)0) return Value_new_ERROR(value,STRAYUNTIL);
        untilpc.token->u.until=*repeatpc;
      }
      break;
    }
    /*}}}*/
    case T_WHILE: /*{{{*/
    {
      struct Pc whilepc=pc;

      if (pass==DECLARE || pass==COMPILE) pushLabel(L_WHILE,&pc);
      ++pc.token;
      if (eval(value,"condition")->type==V_ERROR) return value;
      if (pass==INTERPRET)
      {
        if (Value_isNull(value)) pc=*whilepc.token->u.afterwend;
        Value_destroy(value);
      }
      break;
    }
    /*}}}*/
    case T_WEND: /*{{{*/
    {
      if (pass==DECLARE || pass==COMPILE)
      {
        struct Pc *whilepc;

        if ((whilepc=popLabel(L_WHILE))==(struct Pc*)0) return Value_new_ERROR(value,STRAYWEND);
        *pc.token->u.whilepc=*whilepc;
        ++pc.token;
        *(whilepc->token->u.afterwend)=pc;
      }
      else pc=*pc.token->u.whilepc;
      break;
    }
    /*}}}*/
    case T_WIDTH: /*{{{*/
    {
      int chn=STDCHANNEL,width;

      ++pc.token;
      if (pc.token->type==T_CHANNEL) /*{{{*/
      {
        ++pc.token;
        if (eval(value,"channel")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        chn=value->u.integer;
        Value_destroy(value);
        if (pc.token->type==T_COMMA) ++pc.token;
      }
      /*}}}*/
      if (eval(value,"width")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
      width=value->u.integer;
      Value_destroy(value);
      if (FS_width(chn,width)==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
      break;
    }
    /*}}}*/
    case T_WRITE: /*{{{*/
    {
      int chn=STDCHANNEL;
      int comma=0;

      ++pc.token;
      if (pc.token->type==T_CHANNEL) /*{{{*/
      {
        ++pc.token;
        if (eval(value,"channel")->type==V_ERROR || Value_retype(value,V_INTEGER)->type==V_ERROR) return value;
        chn=value->u.integer;
        Value_destroy(value);
        if (pc.token->type==T_COMMA) ++pc.token;
      }
      /*}}}*/
      while (1)
      {
        if (eval(value,(const char*)0)) 
        {
          if (value->type==V_ERROR) return value;
          if (pass==INTERPRET)
          {
            struct String s;

            String_new(&s);
            if (comma) String_appendChar(&s,',');
            if (FS_putString(chn,Value_toWrite(value,&s))==-1)
            {
              Value_destroy(value);
              return Value_new_ERROR(value,IOERROR,FS_errmsg);
            }
            if (FS_flush(chn)==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
            String_destroy(&s);
          }
          Value_destroy(value);
          comma=1;
        }
        else if (pc.token->type==T_COMMA || pc.token->type==T_SEMICOLON) ++pc.token;
        else break;
      }
      if (pass==INTERPRET)
      {
        FS_putChar(chn,'\n');
        if (FS_flush(chn)==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
      }
      break;
    }
    /*}}}*/
    default: return Value_new_ERROR(value,MISSINGSTATEMENT);
  }
  if (stop) return Value_new_ERROR(value,BREAK);
  else if (pc.token->type==T_COLON && (pc.token+1)->type==T_ELSE) ++pc.token;
  else if ((pc.token->type==T_COLON && (pc.token+1)->type!=T_ELSE) || pc.token->type==T_QUOTE)
  {
    ++pc.token;
    goto more;
  }
  else if ((pass==DECLARE || pass==COMPILE) && pc.token->type!=T_EOL && pc.token->type!=T_ELSE)
  {
    return Value_new_ERROR(value,MISSINGCOLON);
  }
  return Value_new_NIL(value);
}
/*}}}*/

void bas_init(int backslash_colon, int lpfd) /*{{{*/
{
  Token_init(backslash_colon);
  Global_new(&globals);
  Auto_new(&stack);
  Program_new(&program);
  FS_opendev(STDCHANNEL,0,1);
  FS_opendev(LPCHANNEL,-1,lpfd);
}
/*}}}*/
void bas_runFile(const char *runFile) /*{{{*/
{
  struct Value value;
  int dev;

  new();
  if ((dev=FS_openin(runFile))==-1)
  {
    const char *errmsg=FS_errmsg;

    FS_putChars(0,"bas: ");
    FS_putChars(0,errmsg);
    FS_putChar(0,'\n');
  }
  else if (Program_merge(&program,dev,&value))
  {
    struct String s;

    FS_putChars(0,"bas: ");
    String_new(&s);
    Value_toString(&value,&s,' ',-1,0,0,-1,0);
    FS_putString(0,&s);
    String_destroy(&s);
    FS_putChar(0,'\n');
    Value_destroy(&value);
  }
  else
  {
    struct Token line[2];

    line[0].type=T_RUN;
    line[1].type=T_EOL;

    FS_close(dev);
    runline(line);
  }
}
/*}}}*/
void bas_runLine(const char *runLine) /*{{{*/
{
  struct Token *line;

  line=Token_newCode(runLine);
  runline(line+1);
  Token_destroy(line);
}
/*}}}*/
void bas_interpreter(void) /*{{{*/
  {
    FS_putChars(0,"bas " VERSION "\n");
    FS_putChars(0,"Copyright 1999,2000 Michael Haardt.\n");
    FS_putChars(0,"This is free software with ABSOLUTELY NO WARRANTY.\n");
    new();
    while (1)
    {
      struct Token *line;
      struct String s;

      stop=0;
      allowBreak(1);
      FS_putChars(STDCHANNEL,"> ");
      FS_flush(STDCHANNEL);
      String_new(&s);
      if (FS_appendToString(STDCHANNEL,&s,'\n',1)==-1)
      {
        if (stop)
        {
          FS_putChars(STDCHANNEL,"\nBreak\n");
          FS_flush(STDCHANNEL);
          String_destroy(&s);
          continue;
        }
        else
        {
          FS_putChars(STDCHANNEL,FS_errmsg);
          FS_flush(STDCHANNEL);
          String_destroy(&s);
          break;
        }
      }
      if (s.length==0)
      {
        String_destroy(&s);
        break;
      }
      line=Token_newCode(s.character);
      String_destroy(&s);
      if (line->type!=T_EOL)
      {
        if (line->type==T_INTEGER && line->u.integer>0)
        {
          if (program.numbered)
          {
            Program_store(&program,line,line->u.integer);
            runnable=0;
          }
          else
          {
            FS_putChars(STDCHANNEL,"Use `renum' to number program first");
            Token_destroy(line);
          }
        }
        else if (line->type==T_UNNUMBERED)
        {
          runline(line+1);
          Token_destroy(line);
        }
        else
        {
          FS_putChars(STDCHANNEL,"Invalid line\n");
          Token_destroy(line);
        }
      }
      else Token_destroy(line);
    }
    allowBreak(0);
  }
/*}}}*/
void bas_exit(void) /*{{{*/
{
  Auto_destroy(&stack);
  Global_destroy(&globals);
  Program_destroy(&program);
  if (labelStack) free(labelStack);
  FS_closefiles();
  FS_close(LPCHANNEL);
  FS_close(STDCHANNEL);
}
/*}}}*/
